# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - txn item control
"""

import flet as ft


class WuttaTxnItem(ft.Row):
    """
    Control for displaying a transaction line item within main POS
    items list.
    """

    font_size = 24

    def __init__(self, config, row, page=None, *args, **kwargs):
        self.on_reset = kwargs.pop("on_reset", None)

        super().__init__(*args, **kwargs)

        self.config = config
        self.app = config.get_app()
        enum = self.app.enum

        self.row = row

        self.major_style = ft.TextStyle(size=self.font_size, weight=ft.FontWeight.BOLD)

        self.minor_style = ft.TextStyle(size=int(self.font_size * 0.8), italic=True)

        if self.row.row_type in (enum.POS_ROW_TYPE_SELL, enum.POS_ROW_TYPE_OPEN_RING):
            self.build_item_sell()

        # elif self.row.row_type in (self.enum.POS_ROW_TYPE_TENDER,
        #                            self.enum.POS_ROW_TYPE_CHANGE_BACK):
        #     self.build_item_tender()

    def build_item_sell(self):

        self.quantity = ft.TextSpan(style=self.minor_style)
        self.txn_price = ft.TextSpan(style=self.minor_style)

        self.sales_total_style = ft.TextStyle(
            size=self.font_size, weight=ft.FontWeight.BOLD
        )

        self.sales_total = ft.TextSpan(style=self.sales_total_style)

        self.fs_flag = ft.TextSpan(style=self.minor_style)
        self.tax_flag = ft.TextSpan(style=self.minor_style)

        # set initial text display values
        self.refresh(update=False)

        self.controls = [
            ft.Text(
                spans=[
                    ft.TextSpan(f"{self.row.description}", style=self.major_style),
                    ft.TextSpan("× ", style=self.minor_style),
                    self.quantity,
                    ft.TextSpan(" @ ", style=self.minor_style),
                    self.txn_price,
                ],
            ),
            ft.Text(
                spans=[
                    self.fs_flag,
                    self.tax_flag,
                    self.sales_total,
                ],
            ),
        ]
        self.alignment = ft.MainAxisAlignment.SPACE_BETWEEN

    def build_item_tender(self):
        self.controls = [
            ft.Text(
                spans=[
                    ft.TextSpan(f"{self.row.description}", style=self.major_style),
                ],
            ),
            ft.Text(
                spans=[
                    ft.TextSpan(
                        self.app.render_currency(self.row.tender_total),
                        style=self.major_style,
                    ),
                ],
            ),
        ]
        self.alignment = ft.MainAxisAlignment.SPACE_BETWEEN

    def informed_refresh(self, **kwargs):
        pass

    def reset(self, e=None):
        if self.on_reset:
            self.on_reset(e=e)

    def refresh(self, update=True):
        enum = self.app.enum

        if self.row.void:
            self.major_style.color = None
            self.major_style.decoration = ft.TextDecoration.LINE_THROUGH
            self.major_style.weight = None
            self.minor_style.color = None
            self.minor_style.decoration = ft.TextDecoration.LINE_THROUGH
        else:
            self.major_style.color = None
            self.major_style.decoration = None
            self.major_style.weight = ft.FontWeight.BOLD
            self.minor_style.color = None
            self.minor_style.decoration = None

        if self.row.row_type in (enum.POS_ROW_TYPE_SELL, enum.POS_ROW_TYPE_OPEN_RING):
            self.quantity.text = self.app.render_quantity(self.row.quantity)
            self.txn_price.text = self.app.render_currency(self.row.txn_price)
            self.sales_total.text = self.app.render_currency(self.row.sales_total)
            self.fs_flag.text = "FS   " if self.row.foodstamp_eligible else ""
            self.tax_flag.text = f"T{self.row.tax_code}   " if self.row.tax_code else ""

            # if self.line.voided:
            if self.row.void:
                self.sales_total_style.color = None
                self.sales_total_style.decoration = ft.TextDecoration.LINE_THROUGH
                self.sales_total_style.weight = None
            else:
                if (
                    self.row.row_type == enum.POS_ROW_TYPE_SELL
                    and self.row.txn_price_adjusted
                ):
                    self.sales_total_style.color = "orange"
                elif (
                    self.row.row_type == enum.POS_ROW_TYPE_SELL
                    and self.row.cur_price
                    and self.row.cur_price < self.row.reg_price
                ):
                    self.sales_total_style.color = "green"
                else:
                    self.sales_total_style.color = None

                self.sales_total_style.decoration = None
                self.sales_total_style.weight = ft.FontWeight.BOLD

        if update:
            self.update()



# -*- coding: utf-8; -*-
"""
Model definition for Customers
"""

import sqlalchemy as sa
from sqlalchemy import orm

from wuttjamaican.db import model


class Customer(model.Base):
    """
    Technically a customer account, but relates to a person (account holder).
    """

    __tablename__ = "customer"
    __versioned__ = {}
    __wutta_hint__ = {
        "model_title": "Customer",
        "model_title_plural": "Customers",
    }

    uuid = model.uuid_column()

    customer_id = sa.Column(
        sa.String(length=20),
        nullable=False,
        doc="""
        Unique identifier for the customer account.
        """,
    )

    name = sa.Column(
        sa.String(length=100),
        nullable=False,
        doc="""
        Name for the customer account.
        """,
    )

    account_holder_uuid = model.uuid_fk_column("person.uuid", nullable=True)
    account_holder = orm.relationship(
        "Person",
        doc="""
        Reference to the account holder, if applicable.
        """,
    )

    phone_number = sa.Column(
        sa.String(length=20),
        nullable=True,
        doc="""
        Phone number for the customer.
        """,
    )

    email_address = sa.Column(
        sa.String(length=255),
        nullable=True,
        doc="""
        Email address for the customer.
        """,
    )

    def __str__(self):
        return self.name or ""



# -*- coding: utf-8; -*-
"""
Master view for Tenders
"""

from wuttapos.db.model.tenders import Tender

from wuttaweb.views import MasterView


class TenderView(MasterView):
    """
    Master view for Tenders
    """

    model_class = Tender
    model_title = "Tender"
    model_title_plural = "Tenders"

    route_prefix = "tenders"
    url_prefix = "/tenders"

    creatable = True
    editable = True
    deletable = True

    labels = {
        "tender_id": "Tender ID",
    }

    grid_columns = [
        "tender_id",
        "name",
        "is_cash",
        "is_foodstamp",
        "allow_cashback",
        "kick_drawer",
        "active",
    ]

    form_fields = [
        "tender_id",
        "name",
        "notes",
        "is_cash",
        "is_foodstamp",
        "allow_cashback",
        "kick_drawer",
        "active",
    ]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # links
        g.set_link("tender_id")
        g.set_link("name")

    def configure_form(self, form):
        f = form
        super().configure_form(f)

        # notes
        f.set_widget("notes", "notes")


def defaults(config, **kwargs):
    base = globals()

    TenderView = kwargs.get("TenderView", base["TenderView"])
    TenderView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - base lookup control
"""

import flet as ft

from .keyboard import WuttaKeyboard
from wuttapos.terminal.controls.buttons import make_button


class WuttaLookup(ft.Container):

    default_font_size = 40
    font_size = default_font_size * 0.8
    default_button_height_dlg = 80
    disabled_bgcolor = "#aaaaaa"

    long_scroll_delta = 500

    def __init__(self, config, page=None, *args, **kwargs):
        self.on_reset = kwargs.pop("on_reset", None)
        self.show_search = kwargs.pop("show_search", True)
        self.initial_search = kwargs.pop("initial_search", None)
        self.allow_empty_query = kwargs.pop("allow_empty_query", False)
        self.on_select = kwargs.pop("on_select", None)
        self.on_cancel = kwargs.pop("on_cancel", None)

        super().__init__(*args, **kwargs)

        self.config = config
        self.app = config.get_app()
        self.enum = self.app.enum

        # TODO: this feels hacky
        self.mypage = page

        # track current selection
        self.selected_uuid = None
        self.selected_datarow = None

        self.search_results = ft.DataTable(
            columns=[
                ft.DataColumn(self.make_cell_text(text))
                for text in self.get_results_columns()
            ],
            data_row_min_height=50,
        )

        self.no_results = ft.Text(
            "NO RESULTS", size=32, color="red", weight=ft.FontWeight.BOLD, visible=False
        )

        self.select_button = make_button(
            "Select",
            font_size=self.font_size * 0.8,
            height=self.default_button_height_dlg * 0.8,
            width=self.default_button_height_dlg * 1.3,
            on_click=self.select_click,
            disabled=True,
            bgcolor=self.disabled_bgcolor,
        )

        self.up_button = make_button(
            "↑",
            font_size=self.font_size,
            height=self.default_button_height_dlg,
            width=self.default_button_height_dlg,
            on_click=self.up_click,
            on_long_press=self.up_longpress,
        )

        self.down_button = make_button(
            "↓",
            font_size=self.font_size,
            height=self.default_button_height_dlg,
            width=self.default_button_height_dlg,
            on_click=self.down_click,
            on_long_press=self.down_longpress,
        )

        self.search_results_wrapper = ft.Column(
            [
                self.search_results,
                self.no_results,
            ],
            expand=True,
            height=400,
            scroll=ft.ScrollMode.AUTO,
        )

        controls = []

        if self.show_search:

            self.searchbox = ft.TextField(
                "",
                text_size=self.font_size * 0.8,
                on_submit=self.lookup,
                autofocus=True,
                expand=True,
            )

            controls.extend(
                [
                    ft.Row(
                        [
                            ft.Text("SEARCH FOR:"),
                            self.searchbox,
                            make_button(
                                "Lookup",
                                font_size=self.font_size * 0.8,
                                height=self.default_button_height_dlg * 0.8,
                                width=self.default_button_height_dlg * 1.3,
                                on_click=self.lookup,
                                bgcolor="blue",
                            ),
                            make_button(
                                "Reset",
                                font_size=self.font_size * 0.8,
                                height=self.default_button_height_dlg * 0.8,
                                width=self.default_button_height_dlg * 1.3,
                                on_click=self.reset,
                                bgcolor="yellow",
                            ),
                        ],
                    ),
                    ft.Divider(),
                    WuttaKeyboard(
                        self.config,
                        on_keypress=self.keypress,
                        on_long_backspace=self.long_backspace,
                    ),
                ]
            )

        controls.extend(
            [
                ft.Divider(),
                ft.Row(
                    [
                        self.search_results_wrapper,
                        ft.VerticalDivider(),
                        ft.Column(
                            [
                                self.select_button,
                                ft.Row(),
                                ft.Row(),
                                ft.Row(),
                                ft.Row(),
                                ft.Row(),
                                self.up_button,
                                self.down_button,
                                ft.Row(),
                                ft.Row(),
                                ft.Row(),
                                ft.Row(),
                                ft.Row(),
                                make_button(
                                    "Cancel",
                                    font_size=self.font_size * 0.8,
                                    height=self.default_button_height_dlg * 0.8,
                                    width=self.default_button_height_dlg * 1.3,
                                    on_click=self.cancel,
                                ),
                            ],
                        ),
                    ],
                    vertical_alignment=ft.CrossAxisAlignment.START,
                ),
            ]
        )

        self.content = ft.Column(controls=controls)
        self.height = None if self.show_search else 600

    def informed_refresh(self, **kwargs):
        pass

    def reset(self, e=None):
        if self.on_reset:
            self.on_reset(e=e)

    def get_results_columns(self):
        raise NotImplementedError

    def did_mount(self):
        if self.initial_search is not None:
            if self.show_search:
                self.searchbox.value = self.initial_search
            self.initial_search = None  # only do it once
            self.update()
            self.lookup()

    def make_cell_text(self, text):
        return ft.Text(text, size=32)

    def make_cell(self, text):
        return ft.DataCell(self.make_cell_text(text))

    def cancel(self, e):
        if self.on_cancel:
            self.on_cancel(e)

    def keypress(self, key):
        if key == "⏎":
            self.lookup()
        else:
            if key == "⌫":
                self.searchbox.value = self.searchbox.value[:-1]
            else:
                self.searchbox.value += key
            self.searchbox.focus()
            self.update()

    def long_backspace(self):
        self.searchbox.value = self.searchbox.value[:-10]
        self.searchbox.focus()
        self.update()

    def get_results(self, session, entry):
        raise NotImplementedError

    def make_result_row(self, obj):
        return obj

    def lookup(self, e=None):

        if self.show_search:
            entry = self.searchbox.value
            if not entry and not self.allow_empty_query:
                self.searchbox.focus()
                self.update()
                return
        else:
            entry = None

        session = self.app.make_session()
        results = self.get_results(session, entry)

        self.search_results.rows.clear()
        self.selected_uuid = None
        self.select_button.disabled = True
        self.select_button.bgcolor = self.disabled_bgcolor

        if results:
            for obj in results:
                self.search_results.rows.append(
                    ft.DataRow(
                        cells=[
                            self.make_cell(row) for row in self.make_result_row(obj)
                        ],
                        on_select_changed=self.select_changed,
                        data={"uuid": obj["uuid"]},
                    )
                )
            self.no_results.visible = False

        else:
            if self.show_search:
                self.no_results.value = f"NO RESULTS FOR: {entry}"
            else:
                self.no_results.value = "NO RESULTS FOUND"
            self.no_results.visible = True

        if self.show_search:
            self.searchbox.focus()
        self.update()

    def reset(self, e):
        if self.show_search:
            self.searchbox.value = ""
        self.search_results.rows.clear()
        self.no_results.visible = False
        self.selected_uuid = None
        self.selected_datarow = None
        self.select_button.disabled = True
        self.select_button.bgcolor = self.disabled_bgcolor
        if self.show_search:
            self.searchbox.focus()
        self.update()

    def set_selection(self, row):
        if self.selected_datarow:
            self.selected_datarow.color = None

        row.color = ft.colors.BLUE
        self.selected_uuid = row.data["uuid"]
        self.selected_datarow = row

        self.select_button.disabled = False
        self.select_button.bgcolor = "blue"

    def select_changed(self, e):
        if e.data:  # selected
            self.set_selection(e.control)
            self.update()

    def up_click(self, e):

        # select previous row, if selection in progress
        if self.selected_datarow:
            i = self.search_results.rows.index(self.selected_datarow)
            if i > 0:
                self.search_results_wrapper.scroll_to(delta=-48, duration=100)
                self.set_selection(self.search_results.rows[i - 1])
                self.update()
                return

        self.search_results_wrapper.scroll_to(delta=-50, duration=100)
        self.update()

    def up_longpress(self, e):
        self.search_results_wrapper.scroll_to(
            delta=-self.long_scroll_delta, duration=100
        )
        self.update()

    def down_click(self, e):

        # select next row, if selection in progress
        if self.selected_datarow:
            i = self.search_results.rows.index(self.selected_datarow)
            if (i + 1) < len(self.search_results.rows):
                self.search_results_wrapper.scroll_to(delta=48, duration=100)
                self.set_selection(self.search_results.rows[i + 1])
                self.update()
                return

        self.search_results_wrapper.scroll_to(delta=50, duration=100)
        self.update()

    def down_longpress(self, e):
        self.search_results_wrapper.scroll_to(
            delta=self.long_scroll_delta, duration=100
        )
        self.update()

    def select_click(self, e):
        if not self.selected_uuid:
            raise RuntimeError("no record selected?")
        if self.on_select:
            self.on_select(self.selected_uuid)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS enum values
"""

from collections import OrderedDict

from wuttjamaican.enum import *


########################################
# POS Batch Row Type
########################################

POS_ROW_TYPE_SET_CUSTOMER = "set_customer"
POS_ROW_TYPE_SWAP_CUSTOMER = "swap_customer"
POS_ROW_TYPE_DEL_CUSTOMER = "del_customer"
POS_ROW_TYPE_SELL = "sell"
POS_ROW_TYPE_OPEN_RING = "openring"
POS_ROW_TYPE_BADSCAN = "badscan"
POS_ROW_TYPE_BADPRICE = "badprice"
POS_ROW_TYPE_ADJUST_PRICE = "adjust_price"
POS_ROW_TYPE_VOID_LINE = "void_line"
POS_ROW_TYPE_VOID_TXN = "void_txn"
POS_ROW_TYPE_SUSPEND = "suspend"
POS_ROW_TYPE_RESUME = "resume"
POS_ROW_TYPE_TENDER = "tender"
POS_ROW_TYPE_CHANGE_BACK = "change_back"

POS_ROW_TYPE = OrderedDict(
    [
        (POS_ROW_TYPE_SET_CUSTOMER, "set customer"),
        (POS_ROW_TYPE_SWAP_CUSTOMER, "swap customer"),
        (POS_ROW_TYPE_DEL_CUSTOMER, "del customer"),
        (POS_ROW_TYPE_SELL, "sell"),
        (POS_ROW_TYPE_OPEN_RING, "open ring"),
        (POS_ROW_TYPE_BADSCAN, "bad scan"),
        (POS_ROW_TYPE_BADPRICE, "bad price"),
        (POS_ROW_TYPE_ADJUST_PRICE, "adjust price"),
        (POS_ROW_TYPE_VOID_LINE, "void line"),
        (POS_ROW_TYPE_VOID_TXN, "void txn"),
        (POS_ROW_TYPE_SUSPEND, "suspend"),
        (POS_ROW_TYPE_RESUME, "resume"),
        (POS_ROW_TYPE_TENDER, "tender"),
        (POS_ROW_TYPE_CHANGE_BACK, "change back"),
    ]
)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - ten-key menu
"""

import flet as ft

from .base import WuttaMenu


class WuttaTenkeyMenu(WuttaMenu):

    def __init__(self, *args, **kwargs):
        self.simple = kwargs.pop("simple", False)
        self.on_char = kwargs.pop("on_char", None)
        self.on_enter = kwargs.pop("on_enter", None)
        self.on_up_click = kwargs.pop("on_up_click", None)
        self.on_up_longpress = kwargs.pop("on_up_longpress", None)
        self.on_down_click = kwargs.pop("on_down_click", None)
        self.on_down_longpress = kwargs.pop("on_down_longpress", None)

        super().__init__(*args, **kwargs)

    def build_controls(self):

        row1 = [
            self.make_tenkey_button("1"),
            self.make_tenkey_button("2"),
            self.make_tenkey_button("3"),
        ]
        if not self.simple:
            row1.extend(
                [
                    self.make_tenkey_button("-"),
                ]
            )

        row2 = [
            self.make_tenkey_button("4"),
            self.make_tenkey_button("5"),
            self.make_tenkey_button("6"),
        ]
        if not self.simple:
            row2.extend(
                [
                    self.make_tenkey_button("↑", on_long_press=self.up_long_press),
                ]
            )

        row3 = [
            self.make_tenkey_button("7"),
            self.make_tenkey_button("8"),
            self.make_tenkey_button("9"),
        ]
        if not self.simple:
            row3.extend(
                [
                    self.make_tenkey_button("↓", on_long_press=self.down_long_press),
                ]
            )

        row4 = [
            self.make_tenkey_button("0"),
            # self.make_tenkey_button("00"),
            self.make_tenkey_button("."),
        ]
        if self.simple:
            row4.extend(
                [
                    self.make_tenkey_button("⏎"),
                ]
            )
        else:
            row4.extend(
                [
                    self.make_tenkey_button(
                        "ENTER", width=self.default_button_size * 2
                    ),
                ]
            )

        return [
            ft.Row(
                controls=row1,
                spacing=0,
            ),
            ft.Row(
                controls=row2,
                spacing=0,
            ),
            ft.Row(
                controls=row3,
                spacing=0,
            ),
            ft.Row(
                controls=row4,
                spacing=0,
            ),
        ]

    def make_tenkey_button(
        self,
        text,
        width=None,
        on_long_press=None,
    ):
        if not width:
            width = self.default_button_size

        return self.make_button(
            text,
            bgcolor="green",
            width=width,
            on_click=self.tenkey_click,
            on_long_press=on_long_press,
        )

    def tenkey_click(self, e):
        value = e.control.content.value

        if value in ("ENTER", "⏎"):
            if self.on_enter:
                self.on_enter(e)

        elif value == "↑":  # UP
            if self.on_up_click:
                self.on_up_click(e)

        elif value == "↓":  # DOWN
            if self.on_down_click:
                self.on_down_click(e)

        else:  # normal char key
            if self.on_char:
                self.on_char(value)

    def up_long_press(self, e):
        if self.on_up_longpress:
            self.on_up_longpress(e)

    def down_long_press(self, e):
        if self.on_down_longpress:
            self.on_down_longpress(e)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Employment handler
"""

from wuttjamaican.app import GenericHandler


class EmploymentHandler(GenericHandler):
    """
    Base class and default implementation for employment handlers.
    """

    def get_employee(self, obj):
        """
        Returns the Employee associated with the given object, if any.
        """
        model = self.app.model

        if isinstance(obj, model.Employee):
            employee = obj
            return employee

        if person := self.app.get_person(obj):
            if person.employee:
                return person.employee

        return None

    # def make_employee(self, person):
    #     """
    #     Create and return a new employee record.
    #     """
    #     employee = self.model.Employee()
    #     employee.person = person
    #     return employee



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS server views
"""

from wuttaweb.views import essential


def includeme(config):

    # wuttaweb
    essential.defaults(
        config,
        **{
            "wuttaweb.views.common": "wuttapos.server.views.common",
            "wuttaweb.views.settings": "wuttapos.server.views.settings",
        }
    )

    # wuttapos
    config.include("wuttapos.server.views.stores")
    config.include("wuttapos.server.views.terminals")
    config.include("wuttapos.server.views.employees")
    config.include("wuttapos.server.views.tenders")
    config.include("wuttapos.server.views.taxes")
    config.include("wuttapos.server.views.departments")
    config.include("wuttapos.server.views.products")
    config.include("wuttapos.server.views.inventory_adjustments")
    config.include("wuttapos.server.views.customers")
    config.include("wuttapos.server.views.batch.pos")

    # TODO: these should probably live elsewhere?
    config.add_wutta_permission_group("pos", "POS", overwrite=False)

    config.add_wutta_permission(
        "pos", "pos.test_error", "Force error to test error handling"
    )
    config.add_wutta_permission(
        "pos", "pos.ring_sales", "Make transactions (ring sales)"
    )
    config.add_wutta_permission(
        "pos", "pos.override_price", "Override price for any item"
    )
    config.add_wutta_permission(
        "pos", "pos.del_customer", "Remove customer from current transaction"
    )
    # config.add_wutta_permission('pos', 'pos.resume',
    #                                "Resume previously-suspended transaction")
    config.add_wutta_permission("pos", "pos.toggle_training", "Start/end training mode")
    config.add_wutta_permission("pos", "pos.suspend", "Suspend current transaction")
    config.add_wutta_permission(
        "pos", "pos.swap_customer", "Swap customer for current transaction"
    )
    config.add_wutta_permission("pos", "pos.void_txn", "Void current transaction")



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for app settings
"""

import glob
import os
import logging

from wuttaweb.views import settings as base


log = logging.getLogger(__name__)


class AppInfoView(base.AppInfoView):
    """
    We override this view to add the :meth:`install_sample_data()` route.

    See also parent docs, :class:`~wuttaweb:wuttaweb.views.settings.AppInfoView`
    """

    def install_sample_data(self):
        """
        Special view to install sample data.  Probably just temporary until
        something better can be worked out...

        Requires root access; imports (create only) from small CSV
        files within the WuttaPOS code base.
        """

        # only root gets to do this
        if not self.request.is_root:
            raise self.forbidden()

        # we can't (yet?) tell the importer to import "everything"
        # because it will think that means "all models" as opposed to
        # "all files" found, and that raises error for missing files.
        # so we first inspect the folder to see what's there.
        models = []
        samples = self.app.resource_path("wuttapos:sample-data")
        # nb. should use root_dir param for glob() but requires python 3.10
        for path in glob.glob(os.path.join(samples, "*.csv")):
            name, ext = os.path.splitext(os.path.basename(path))
            models.append(name)

        # import all CSV files found in wuttapos/sample-data, but only
        # allow new records to be created, never update/delete
        try:
            handler = self.app.get_import_handler("import.to_wutta.from_csv")
            models = [name for name in models if name in handler.importers]
            kw = dict(input_file_path=samples, allow_update=False, allow_delete=False)
            handler.process_data(*models, **kw)
        except Exception as err:
            log.warning("failed to install sample data", exc_info=True)
            self.request.session.flash(
                f"Install failed: {self.app.render_error(err)}", "error"
            )
        else:
            self.request.session.flash("Sample data has been installed.")

        return self.redirect(self.get_index_url())

    @classmethod
    def defaults(cls, config):  # pylint: disable=empty-docstring
        """ """
        cls._defaults(config)
        cls._appinfo_defaults(config)
        cls._wuttapos_defaults(config)

    @classmethod
    def _wuttapos_defaults(cls, config):
        route_prefix = cls.get_route_prefix()
        permission_prefix = cls.get_permission_prefix()
        url_prefix = cls.get_url_prefix()

        # install sample data
        config.add_route(
            f"{route_prefix}.install_sample_data",
            f"{url_prefix}/install-sample-data",
            request_method="POST",
        )
        config.add_view(
            cls,
            attr="install_sample_data",
            route_name=f"{route_prefix}.install_sample_data",
        )


def defaults(config, **kwargs):  # pylint: disable=missing-function-docstring
    local = globals()

    AppInfoView = kwargs.get("AppInfoView", local["AppInfoView"])

    base.defaults(config, **{"AppInfoView": AppInfoView})


def includeme(config):  # pylint: disable=missing-function-docstring
    defaults(config)



"""add Taxes

Revision ID: 7067ef686eb0
Revises: b026e4f5c5bc
Create Date: 2026-01-02 12:15:45.279817

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "7067ef686eb0"
down_revision: Union[str, None] = "b026e4f5c5bc"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # tax
    op.create_table(
        "tax",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("tax_id", sa.String(length=20), nullable=False),
        sa.Column("name", sa.String(length=50), nullable=False),
        sa.Column("rate", sa.Numeric(precision=7, scale=5), nullable=False),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_tax")),
    )
    op.create_table(
        "tax_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column("tax_id", sa.String(length=20), autoincrement=False, nullable=True),
        sa.Column("name", sa.String(length=50), autoincrement=False, nullable=True),
        sa.Column(
            "rate", sa.Numeric(precision=7, scale=5), autoincrement=False, nullable=True
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint("uuid", "transaction_id", name=op.f("pk_tax_version")),
    )
    op.create_index(
        op.f("ix_tax_version_end_transaction_id"),
        "tax_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_tax_version_operation_type"),
        "tax_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        op.f("ix_tax_version_transaction_id"),
        "tax_version",
        ["transaction_id"],
        unique=False,
    )


def downgrade() -> None:

    # tax
    op.drop_table("batch_pos_row")
    op.drop_table("batch_pos")
    op.drop_index(op.f("ix_tax_version_transaction_id"), table_name="tax_version")
    op.drop_index(op.f("ix_tax_version_operation_type"), table_name="tax_version")
    op.drop_index(op.f("ix_tax_version_end_transaction_id"), table_name="tax_version")
    op.drop_table("tax_version")
    op.drop_table("tax")



"""add Departments

Revision ID: 148d701e4ac7
Revises:
Create Date: 2026-01-01 17:26:20.695393

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "148d701e4ac7"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = ("wuttapos",)
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # department
    op.create_table(
        "department",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("department_id", sa.String(length=20), nullable=False),
        sa.Column("name", sa.String(length=100), nullable=False),
        sa.Column("for_products", sa.Boolean(), nullable=False),
        sa.Column("for_personnel", sa.Boolean(), nullable=False),
        sa.Column("exempt_from_gross_sales", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_department")),
    )
    op.create_table(
        "department_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "department_id", sa.String(length=20), autoincrement=False, nullable=True
        ),
        sa.Column("name", sa.String(length=100), autoincrement=False, nullable=True),
        sa.Column("for_products", sa.Boolean(), autoincrement=False, nullable=True),
        sa.Column("for_personnel", sa.Boolean(), autoincrement=False, nullable=True),
        sa.Column(
            "exempt_from_gross_sales", sa.Boolean(), autoincrement=False, nullable=True
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_department_version")
        ),
    )
    op.create_index(
        op.f("ix_department_version_end_transaction_id"),
        "department_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_department_version_operation_type"),
        "department_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        op.f("ix_department_version_transaction_id"),
        "department_version",
        ["transaction_id"],
        unique=False,
    )


def downgrade() -> None:

    # department
    op.drop_index(
        op.f("ix_department_version_transaction_id"), table_name="department_version"
    )
    op.drop_index(
        op.f("ix_department_version_operation_type"), table_name="department_version"
    )
    op.drop_index(
        op.f("ix_department_version_end_transaction_id"),
        table_name="department_version",
    )
    op.drop_table("department_version")
    op.drop_table("department")



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - transaction lookup control
"""

from .lookup import WuttaLookup


class WuttaTransactionLookup(WuttaLookup):

    def __init__(self, *args, **kwargs):

        # nb. this forces first query
        kwargs.setdefault("initial_search", True)

        # TODO: how to deal with 'modes'
        self.mode = kwargs.pop("mode", None)
        if not self.mode:
            raise ValueError("must specify mode")
        if self.mode != "resume":
            raise ValueError("only 'resume' mode is supported")

        kwargs.setdefault("show_search", False)

        super().__init__(*args, **kwargs)

    def get_results_columns(self):
        return [
            "Created",
            "Terminal",
            "Txn ID",
            "Cashier",
            "Customer",
            "Balance",
        ]

    def get_results(self, session, entry):
        model = self.app.model

        # TODO: how to deal with 'modes'
        assert self.mode == "resume"
        training = bool(self.mypage.session.get("training"))
        query = (
            session.query(model.POSBatch)
            .filter(model.POSBatch.status_code == model.POSBatch.STATUS_SUSPENDED)
            .filter(model.POSBatch.executed == None)
            .filter(model.POSBatch.training_mode == training)
            .order_by(model.POSBatch.created.desc())
        )

        transactions = []
        for batch in query:
            transactions.append(
                {
                    "uuid": batch.uuid,
                    "created": self.app.render_datetime(batch.created),
                    "terminal": batch.terminal.terminal_id,
                    "txnid": batch.id_str,
                    "cashier": batch.cashier.name,
                    "customer": batch.customer.name,
                    "balance": self.app.render_currency(batch.get_balance()),
                }
            )
        return transactions

    def make_result_row(self, txn):
        return [
            txn["created"],
            txn["terminal"],
            txn["txnid"],
            txn["cashier"],
            txn["customer"],
            txn["balance"],
        ]



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Model for POS Batches
"""

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.orderinglist import ordering_list

from wuttjamaican.db import model


class POSBatch(model.BatchMixin, model.Base):
    """
    Hopefully generic batch used for entering new purchases into the system, etc.?
    """

    __tablename__ = "batch_pos"
    __wutta_hint__ = {
        "model_title": "POS Batch",
        "model_title_plural": "POS Batches",
    }

    batch_type = "pos"

    STATUS_OK = 1
    STATUS_SUSPENDED = 2

    STATUS = {
        STATUS_OK: "ok",
        STATUS_SUSPENDED: "suspended",
    }

    store_uuid = model.uuid_fk_column("store.uuid", nullable=False)
    store = orm.relationship(
        "Store",
        doc="""
        Reference to the store where the transaction ocurred.
        """,
    )

    terminal_uuid = model.uuid_fk_column("terminal.uuid", nullable=False)
    terminal = orm.relationship(
        "Terminal",
        doc="""
        Reference to the terminal where the transaction ocurred.
        """,
    )

    # receipt_number = sa.Column(sa.String(length=20), nullable=True, doc="""
    # Receipt number for the transaction, if known.
    # """)

    cashier_uuid = model.uuid_fk_column("employee.uuid", nullable=False)
    cashier = orm.relationship(
        "Employee",
        doc="""
        Reference to the employee who acted as cashier.
        """,
    )

    customer_uuid = model.uuid_fk_column("customer.uuid", nullable=True)
    customer = orm.relationship(
        "Customer",
        doc="""
        Reference to the customer account for the transaction.
        """,
    )

    customer_is_member = sa.Column(
        sa.Boolean(),
        nullable=True,
        doc="""
    Flag indicating the customer was a "member" at time of sale.
    """,
    )

    customer_is_employee = sa.Column(
        sa.Boolean(),
        nullable=True,
        doc="""
    Flag indicating the customer was an employee at time of sale.
    """,
    )

    # shopper_number = sa.Column(sa.String(length=20), nullable=True, doc="""
    # Number of the shopper account for the transaction, if applicable.
    # """)

    # shopper_name = sa.Column(sa.String(length=255), nullable=True, doc="""
    # Name of the shopper account for the transaction, if applicable.
    # """)

    # shopper_uuid = sa.Column(sa.String(length=32), nullable=True)
    # shopper = orm.relationship(
    #     'CustomerShopper',
    #     doc="""
    #     Reference to the shopper account for the transaction.
    #     """)

    sales_total = sa.Column(
        sa.Numeric(precision=9, scale=2),
        nullable=True,
        doc="""
    Sales total for the transaction.
    """,
    )

    fs_total = sa.Column(
        sa.Numeric(precision=9, scale=2),
        nullable=True,
        doc="""
    Portion of the sales total which is foodstamp-eligible.
    """,
    )

    tax_total = sa.Column(
        sa.Numeric(precision=9, scale=2),
        nullable=True,
        doc="""
    Tax total for the transaction.
    """,
    )

    fs_tender_total = sa.Column(
        sa.Numeric(precision=9, scale=2),
        nullable=True,
        doc="""
    Foodstamp tender total for the transaction.
    """,
    )

    tender_total = sa.Column(
        sa.Numeric(precision=9, scale=2),
        nullable=True,
        doc="""
    Tender total for the transaction.
    """,
    )

    void = sa.Column(
        sa.Boolean(),
        nullable=False,
        default=False,
        doc="""
    Flag indicating if the transaction was voided.
    """,
    )

    training_mode = sa.Column(
        sa.Boolean(),
        nullable=False,
        default=False,
        doc="""
    Flag indicating if the transaction was rang in training mode,
    i.e. not real / should not go on the books.
    """,
    )

    def get_balance(self):
        return (
            (self.sales_total or 0) + (self.tax_total or 0) + (self.tender_total or 0)
        )

    def get_fs_balance(self):
        return (self.fs_total or 0) + (self.fs_tender_total or 0)


class POSBatchRow(model.BatchRowMixin, model.Base):
    """
    Row of data within a POS batch.
    """

    __tablename__ = "batch_pos_row"
    __batch_class__ = POSBatch

    STATUS_OK = 1

    STATUS = {
        STATUS_OK: "ok",
    }

    modified_by_uuid = model.uuid_fk_column("user.uuid", nullable=False)
    modified_by = orm.relationship(
        "User",
        doc="""
        Reference to the user who added this row to the batch.
        """,
    )

    row_type = sa.Column(
        sa.String(length=20),
        nullable=False,
        doc="""
    Type of item represented by this row, e.g. "item" or "return" or
    "tender" etc.

    .. todo::
       need to figure out how to manage/track POSBatchRow.row_type
    """,
    )

    item_entry = sa.Column(
        sa.String(length=20),
        nullable=True,
        doc="""
    Raw/original entry value for the item, if applicable.
    """,
    )

    description = sa.Column(
        sa.String(length=100),
        nullable=True,
        doc="""
    Description for the row.
    """,
    )

    product_uuid = model.uuid_fk_column("product.uuid", nullable=True)
    product = orm.relationship(
        "Product",
        doc="""
        Reference to the associated product, if applicable.
        """,
    )

    # department_uuid = model.uuid_fk_column("department.uuid", nullable=True)
    # department = orm.relationship(
    #     "Department",
    #     doc="""
    #     Reference to the associated department, if applicable.
    #     """,
    # )

    # subdepartment_number = sa.Column(
    #     sa.Integer(),
    #     nullable=True,
    #     doc="""
    # Number of the subdepartment to which the product belongs.
    # """,
    # )

    # subdepartment_name = sa.Column(
    #     sa.String(length=30),
    #     nullable=True,
    #     doc="""
    # Name of the subdepartment to which the product belongs.
    # """,
    # )

    foodstamp_eligible = sa.Column(
        sa.Boolean(),
        nullable=True,
        doc="""
    Indicates the item is eligible for purchase with food stamps
    or equivalent.
    """,
    )

    sold_by_weight = sa.Column(
        sa.Boolean(),
        nullable=True,
        doc="""
    Flag indicating the item is sold by weight.
    """,
    )

    quantity = sa.Column(
        sa.Numeric(precision=8, scale=2),
        nullable=True,
        doc="""
    Quantity for the item.
    """,
    )

    cost = sa.Column(
        sa.Numeric(precision=8, scale=3),
        nullable=True,
        doc="""
    Internal cost for the item sold.

    NOTE: this may need to change at some point, hence the "generic"
    naming so far.  would we need to record multiple kinds of costs?
    """,
    )

    reg_price = sa.Column(
        sa.Numeric(precision=8, scale=3),
        nullable=True,
        doc="""
    Regular price for the item.
    """,
    )

    cur_price = sa.Column(
        sa.Numeric(precision=8, scale=3),
        nullable=True,
        doc="""
    Current price for the item.
    """,
    )

    cur_price_type = sa.Column(
        sa.Integer(),
        nullable=True,
        doc="""
    Type code for the current price, if applicable.
    """,
    )

    cur_price_start = sa.Column(
        sa.DateTime(),
        nullable=True,
        doc="""
    Start date for current price, if applicable.
    """,
    )

    cur_price_end = sa.Column(
        sa.DateTime(),
        nullable=True,
        doc="""
    End date for current price, if applicable.
    """,
    )

    txn_price = sa.Column(
        sa.Numeric(precision=8, scale=3),
        nullable=True,
        doc="""
    Actual price paid for the item.
    """,
    )

    txn_price_adjusted = sa.Column(
        sa.Boolean(),
        nullable=True,
        doc="""
    Flag indicating the actual price was manually adjusted.
    """,
    )

    sales_total = sa.Column(
        sa.Numeric(precision=9, scale=2),
        nullable=True,
        doc="""
    Sales total for the item.
    """,
    )

    tax_code = sa.Column(
        sa.String(length=30),
        nullable=True,
        doc="""
    Unique "code" for the item tax rate, if applicable.
    """,
    )

    # tax_uuid = model.uuid_fk_column("tax.uuid", nullable=True)
    # tax = orm.relationship(
    #     "Tax",
    #     doc="""
    #     Reference to the associated tax, if applicable.
    #     """,
    # )

    tender_total = sa.Column(
        sa.Numeric(precision=9, scale=2),
        nullable=True,
        doc="""
    Tender total for the item.
    """,
    )

    # tender_uuid = model.uuid_fk_column("tender.uuid", nullable=True)
    # tender = orm.relationship(
    #     "Tender",
    #     doc="""
    #     Reference to the associated tender, if applicable.
    #     """,
    # )

    void = sa.Column(
        sa.Boolean(),
        nullable=False,
        default=False,
        doc="""
    Flag indicating the line item was voided.
    """,
    )



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Products Handler
"""

# import decimal
import logging
import uuid as _uuid

from sqlalchemy import orm

from wuttjamaican.app import GenericHandler


log = logging.getLogger(__name__)


class ProductsHandler(GenericHandler):
    """
    Base class and default implementation for product handlers.

    A products handler of course should get the final say in how products are
    handled.  This means everything from pricing, to whether or not a
    particular product can be deleted, etc.
    """

    def locate_product_for_entry(
        self, session, entry, include_not_for_sale=False, **kwargs
    ):
        """
        This method aims to provide sane default logic for locating a
        :class:`~rattail.db.model.products.Product` record for the
        given "entry" value.

        The default logic here will try to honor the "configured"
        product key field, and prefer that when attempting the lookup.

        :param session: Reference to current DB session.

        :param entry: Value to use for lookup.  This is most often a
           simple string, but the method can handle a few others.  For
           instance it is common to read values from a spreadsheet,
           and sometimes those come through as integers etc.  If this
           value is a :class:`~rattail.gpc.GPC` instance, special
           logic may be used for the lookup.

        :param lookup_fields: Optional list of fields to use for
           lookup.  The default value is ``['uuid', '_product_key_']``
           which means to lookup by UUID as well as "product key"
           field, which is configurable.  You can include any of the
           following in ``lookup_fields``:

           * ``uuid``
           * ``_product_key_`` - :meth:`locate_product_for_key`
           * ``upc`` - :meth:`locate_product_for_upc`
           * ``item_id`` - :meth:`locate_product_for_item_id`
           * ``scancode`` - :meth:`locate_product_for_scancode`
           * ``vendor_code`` - :meth:`locate_product_for_vendor_code`
           * ``alt_code`` - :meth:`locate_product_for_alt_code`

        :param include_not_for_sale: Optional flag to include items
           which are "not for sale" in the search results.

        :returns: First :class:`~rattail.db.model.products.Product`
           instance found if there was a match; otherwise ``None``.
        """
        model = self.app.model
        if not entry:
            return

        # figure out which fields we should match on
        # TODO: let config declare default lookup_fields
        lookup_fields = kwargs.pop(
            "lookup_fields",
            [
                "uuid",
                "product_id",
            ],
        )

        kwargs["include_not_for_sale"] = include_not_for_sale

        # try to locate product by uuid before other, more specific key
        if "uuid" in lookup_fields:
            if isinstance(entry, (_uuid.UUID, str)):
                if product := session.get(model.Product, entry):
                    return product
                # # TODO: should we ever return deleted product?
                # if product and not product.deleted:
                #     if include_not_for_sale or not product.not_for_sale:
                #         return product

        lookups = {
            "uuid": None,
            "product_id": self.locate_product_for_id,
            # "upc": self.locate_product_for_upc,
            # "item_id": self.locate_product_for_item_id,
            # "scancode": self.locate_product_for_scancode,
            # "vendor_code": self.locate_product_for_vendor_code,
            # "alt_code": self.locate_product_for_alt_code,
        }

        for field in lookup_fields:
            if field in lookups:
                lookup = lookups[field]
                if lookup:
                    product = lookup(session, entry, **kwargs)
                    if product:
                        return product
            else:
                log.warning("unknown lookup field: %s", field)

    def locate_product_for_id(
        self,
        session,
        entry,
        # include_not_for_sale=False,
        # include_deleted=False,
        **kwargs,
    ):
        """
        Locate the product which matches the given item ID.

        This will do a lookup on the
        :attr:`rattail.db.model.products.Product.item_id` field only.

        Note that instead of calling this method directly, you might
        consider calling :meth:`locate_product_for_key` instead.

        :param session: Current session for Rattail DB.

        :param entry: Item ID value as string.

        :param include_not_for_sale: Optional flag to include items
           which are "not for sale" in the search results.

        :param include_deleted: Whether "deleted" products should ever
           match (and be returned).

        :returns: First :class:`~rattail.db.model.products.Product`
           instance found if there was a match; otherwise ``None``.
        """
        if not entry:
            return

        # assume entry is string
        entry = str(entry)

        model = self.app.model
        products = session.query(model.Product)
        # if not include_deleted:
        #     products = products.filter(model.Product.deleted == False)
        # if not include_not_for_sale:
        #     products = products.filter(model.Product.not_for_sale == False)

        try:
            return products.filter(model.Product.product_id == entry).one()
        except orm.exc.NoResultFound:
            return None

    def search_products(self, session, entry, **kwargs):
        """
        Perform a product search across multiple fields, and return
        results as JSON data rows.
        """
        model = self.app.model
        final_results = []

        # first we'll attempt "lookup" logic..

        lookup_fields = kwargs.get(
            "lookup_fields",
            [
                # "_product_key_",
                "product_id"
            ],
        )

        if lookup_fields:
            product = self.locate_product_for_entry(
                session, entry, lookup_fields=lookup_fields
            )
            if product:
                final_results.append(product)

        # then we'll attempt "search" logic..

        search_fields = kwargs.get(
            "search_fields",
            [
                "product_id",
                "brand",
                "description",
                "size",
            ],
        )

        searches = {
            "product_id": self.search_products_for_product_id,
            "brand": self.search_products_for_brand,
            "description": self.search_products_for_description,
            "size": self.search_products_for_size,
        }

        for field in search_fields:
            if field in searches:
                search = searches[field]
                if search:
                    products = search(session, entry, **kwargs)
                    final_results.extend(products)
            else:
                log.warning("unknown search field: %s", field)

        return [self.normalize_product(c) for c in final_results]

    def search_products_for_product_id(
        self,
        session,
        entry,
        # include_not_for_sale=False,
    ):
        """
        Search for products where the
        :attr:`~rattail.db.model.products.Product.item_id` contains
        the given value.

        :param entry: Search term.

        :param include_not_for_sale: Optional flag to include items
           which are "not for sale" in the search results.

        :returns: List of products matching the search.
        """
        model = self.app.model
        entry = entry.lower()

        products = session.query(model.Product).filter(
            model.Product.product_id.ilike(f"%{entry}%")
        )

        # if not include_not_for_sale:
        #     products = products.filter(model.Product.not_for_sale == False)

        return products.all()

    def search_products_for_brand(
        self,
        session,
        entry,
        # include_not_for_sale=False,
    ):
        """
        Search for products where the brand
        :attr:`~rattail.db.model.products.Brand.name` contains the
        given value.

        :param entry: Search term.

        :param include_not_for_sale: Optional flag to include items
           which are "not for sale" in the search results.

        :returns: List of products matching the search.
        """
        model = self.app.model
        entry = entry.lower()

        products = session.query(model.Product).filter(
            model.Product.brand_name.ilike(f"%{entry}%")
        )

        # if not include_not_for_sale:
        #     products = products.filter(model.Product.not_for_sale == False)

        return products.all()

    def search_products_for_description(
        self,
        session,
        entry,
        # include_not_for_sale=False,
    ):
        """
        Search for products where the
        :attr:`~rattail.db.model.products.Product.description`
        contains the given value.

        :param entry: Search term.

        :param include_not_for_sale: Optional flag to include items
           which are "not for sale" in the search results.

        :returns: List of products matching the search.
        """
        model = self.app.model
        entry = entry.lower()

        products = session.query(model.Product).filter(
            model.Product.description.ilike(f"%{entry}%")
        )

        # if not include_not_for_sale:
        #     products = products.filter(model.Product.not_for_sale == False)

        return products.all()

    def search_products_for_size(
        self,
        session,
        entry,
        # include_not_for_sale=False,
    ):
        """
        Search for products where the
        :attr:`~rattail.db.model.products.Product.size` contains the
        given value.

        :param entry: Search term.

        :param include_not_for_sale: Optional flag to include items
           which are "not for sale" in the search results.

        :returns: List of products matching the search.
        """
        model = self.app.model
        entry = entry.lower()

        products = session.query(model.Product).filter(
            model.Product.size.ilike(f"%{entry}%")
        )

        # if not include_not_for_sale:
        #     products = products.filter(model.Product.not_for_sale == False)

        return products.all()

    def normalize_product(self, product, fields=None):
        """
        Normalize the given product to a JSON-serializable dict.
        """
        data = {
            "uuid": product.uuid,
            "product_id": product.product_id,
            "description": product.description,
            "size": product.size,
            "_str": str(product),
        }

        if not fields:
            fields = [
                "brand_name",
                "full_description",
                "department_name",
                "unit_price_display",
            ]

        if "brand_name" in fields:
            data["brand_name"] = product.brand_name

        if "full_description" in fields:
            data["full_description"] = product.full_description

        if "department_name" in fields:
            data["department_name"] = (
                product.department.name if product.department else None
            )

        if "unit_price_display" in fields:
            data["unit_price_display"] = self.app.render_currency(
                product.unit_price_reg
            )

        # if "vendor_name" in fields:
        #     vendor = product.cost.vendor if product.cost else None
        #     data["vendor_name"] = vendor.name if vendor else None

        # if "costs" in fields:
        #     costs = []
        #     for cost in product.costs:
        #         costs.append(
        #             {
        #                 "uuid": cost.uuid,
        #                 "vendor_uuid": cost.vendor_uuid,
        #                 "vendor_name": cost.vendor.name,
        #                 "preference": cost.preference,
        #                 "code": cost.code,
        #                 "case_size": cost.case_size,
        #                 "case_cost": cost.case_cost,
        #                 "unit_cost": cost.unit_cost,
        #             }
        #         )
        #     data["costs"] = costs

        # current_price = None
        # if not product.not_for_sale:

        #     margin_fields = [
        #         "true_margin",
        #         "true_margin_display",
        #     ]
        #     if any([f in fields for f in margin_fields]):
        #         if product.volatile:
        #             data["true_margin"] = product.volatile.true_margin
        #             data["true_margin_display"] = self.app.render_percent(
        #                 product.volatile.true_margin, places=2
        #             )

        #     current_fields = [
        #         "current_price",
        #         "current_price_display",
        #         "current_ends",
        #         "current_ends_display",
        #     ]
        #     if any([f in fields for f in current_fields]):
        #         current_price = product.current_price
        #         if current_price:
        #             if current_price.price:
        #                 data["current_price"] = float(current_price.price)
        #             data["current_price_display"] = self.render_price(current_price)
        #             current_ends = current_price.ends
        #             if current_ends:
        #                 current_ends = self.app.localtime(
        #                     current_ends, from_utc=True
        #                 ).date()
        #                 data["current_ends"] = str(current_ends)
        #                 data["current_ends_display"] = self.app.render_date(
        #                     current_ends
        #                 )

        #     sale_fields = [
        #         "sale_price",
        #         "sale_price_display",
        #         "sale_ends",
        #         "sale_ends_display",
        #     ]
        #     if any([f in fields for f in sale_fields]):
        #         sale_price = product.sale_price
        #         if sale_price:
        #             if sale_price.price:
        #                 data["sale_price"] = float(sale_price.price)
        #             data["sale_price_display"] = self.render_price(sale_price)
        #             sale_ends = sale_price.ends
        #             if sale_ends:
        #                 sale_ends = self.app.localtime(sale_ends, from_utc=True).date()
        #                 data["sale_ends"] = str(sale_ends)
        #                 data["sale_ends_display"] = self.app.render_date(sale_ends)

        #     tpr_fields = [
        #         "tpr_price",
        #         "tpr_price_display",
        #         "tpr_ends",
        #         "tpr_ends_display",
        #     ]
        #     if any([f in fields for f in tpr_fields]):
        #         tpr_price = product.tpr_price
        #         if tpr_price:
        #             if tpr_price.price:
        #                 data["tpr_price"] = float(tpr_price.price)
        #             data["tpr_price_display"] = self.render_price(tpr_price)
        #             tpr_ends = tpr_price.ends
        #             if tpr_ends:
        #                 tpr_ends = self.app.localtime(tpr_ends, from_utc=True).date()
        #                 data["tpr_ends"] = str(tpr_ends)
        #                 data["tpr_ends_display"] = self.app.render_date(tpr_ends)

        if "case_size" in fields:
            data["case_size"] = self.app.render_quantity(self.get_case_size(product))

        # if "case_price" in fields or "case_price_display" in fields:
        #     case_price = None
        #     if product.regular_price and product.regular_price is not None:
        #         case_size = self.get_case_size(product)
        #         # use "current" price if there is one, else normal unit price
        #         unit_price = product.regular_price.price
        #         if current_price:
        #             unit_price = current_price.price
        #         case_price = (case_size or 1) * unit_price
        #         case_price = case_price.quantize(decimal.Decimal("0.01"))
        #     data["case_price"] = str(case_price) if case_price is not None else None
        #     data["case_price_display"] = self.app.render_currency(case_price)

        # if "uom_choices" in fields:
        #     data["uom_choices"] = self.get_uom_choices(product)

        return data

    def get_case_size(self, product):
        """
        Return the effective case size for the given product.
        """
        if product.case_size is not None:
            return product.case_size

        # cost = product.cost
        # if cost:
        #     return cost.case_size

        return None



# -*- coding: utf-8; -*-
"""
Model definition for Products
"""

import sqlalchemy as sa
from sqlalchemy import orm

from wuttjamaican.db import model
from wuttjamaican.db.util import UUID


class Product(model.Base):
    """
    Represents an item for sale (usually).
    """

    __tablename__ = "product"
    __versioned__ = {}
    __wutta_hint__ = {
        "model_title": "Product",
        "model_title_plural": "Products",
    }

    uuid = model.uuid_column()

    product_id = sa.Column(
        sa.String(length=20),
        nullable=False,
        doc="""
        Unique identifier for the product.
        """,
    )

    brand_name = sa.Column(
        sa.String(length=100),
        nullable=True,
        doc="""
        Brand name for the product, if applicable.
        """,
    )

    description = sa.Column(
        sa.String(length=255),
        nullable=False,
        doc="""
        Description of the product.
        """,
    )

    size = sa.Column(
        sa.String(length=30),
        nullable=True,
        doc="""
        Size of the product.
        """,
    )

    sold_by_weight = sa.Column(
        sa.Boolean(),
        nullable=False,
        doc="""
        Indicates the item is sold by weight, vs. by single units.
        """,
    )

    department_uuid = model.uuid_fk_column("department.uuid", nullable=False)
    department = orm.relationship(
        "Department",
        doc="""
        Department to which the product belongs.
        """,
    )

    special_order = sa.Column(
        sa.Boolean(),
        nullable=True,
        doc="""
        Indicates the item is not normally carried, must be ordered specially.
        """,
    )

    case_size = sa.Column(
        sa.Numeric(precision=9, scale=4),
        nullable=True,
        doc="""
        Number of units in a case for this item, if applicable.
        """,
    )

    unit_cost = sa.Column(
        sa.Numeric(precision=9, scale=5),
        nullable=True,
        doc="""
        Current cost (from the vendor, to the retailer) for one unit of the item.
        """,
    )

    unit_price_reg = sa.Column(
        sa.Numeric(precision=8, scale=3),
        nullable=True,
        doc="""
        Regular price (to the customer) for one unit of product.
        """,
    )

    notes = sa.Column(
        sa.Text(),
        nullable=True,
        doc="""
        Arbitrary notes for the product.
        """,
    )

    inventory = orm.relationship(
        "ProductInventory",
        doc="""
        Reference to the live inventory record for this product.
        """,
        uselist=False,
        back_populates="product",
        cascade="all, delete-orphan",
    )

    @property
    def full_description(self):
        fields = [self.brand_name or "", self.description or "", self.size or ""]
        fields = [f.strip() for f in fields if f.strip()]
        return " ".join(fields)

    def __str__(self):
        return self.full_description


class ProductInventory(model.Base):
    """
    Contains the live inventory counts for products.
    """

    __tablename__ = "product_inventory"
    __wutta_hint__ = {
        "model_title": "Product Inventory",
        "model_title_plural": "Product Inventory",
    }

    uuid = model.uuid_column()

    product_uuid = model.uuid_fk_column("product.uuid", nullable=False)
    product = orm.relationship(
        "Product",
        doc="""
        Reference to the product.
        """,
        back_populates="inventory",
    )

    on_hand = sa.Column(
        sa.Numeric(precision=9, scale=4),
        nullable=True,
        doc="""
        Unit quantity of product which is currently on hand.
        """,
    )

    on_order = sa.Column(
        sa.Numeric(precision=9, scale=4),
        nullable=True,
        doc="""
        Unit quantity of product which is currently on order.
        """,
    )

    adjustments = orm.relationship(
        "InventoryAdjustment",
        back_populates="inventory",
        cascade="all, delete-orphan",
    )

    def __str__(self):
        return str(self.product or "")


class InventoryAdjustmentType(model.Base):
    """
    Possible types of inventory adjustments.
    """

    __tablename__ = "inventory_adjustment_type"
    __versioned__ = {}
    __wutta_hint__ = {
        "model_title": "Inventory Adjustment Type",
        "model_title_plural": "Inventory Adjustment Types",
    }

    uuid = model.uuid_column()

    type_code = sa.Column(
        sa.Integer(),
        nullable=False,
        unique=True,
        doc="""
        Code indicating the type of inventory adjustment.
        """,
    )

    name = sa.Column(
        sa.String(length=100),
        nullable=False,
        doc="""
        Name for the adjustment type.
        """,
    )

    def __str__(self):
        return self.name or ""


class InventoryAdjustment(model.Base):
    """
    Represents any adjustment to inventory.
    """

    __tablename__ = "inventory_adjustment"
    __wutta_hint__ = {
        "model_title": "Inventory Adjustment",
        "model_title_plural": "Inventory Adjustments",
    }

    uuid = model.uuid_column()

    inventory_uuid = model.uuid_fk_column("product_inventory.uuid", nullable=True)
    inventory = orm.relationship(
        "ProductInventory",
        doc="""
        Reference to the product inventory record.
        """,
        back_populates="adjustments",
    )

    adjusted = sa.Column(
        sa.DateTime(),
        nullable=False,
        doc="""
        Date and time (in UTC) when the adjustment occurred.
        """,
    )

    effective_date = sa.Column(
        sa.Date(),
        nullable=False,
        doc="""
        Effective date (in local time zone) for the adjustment.
        """,
    )

    adjustment_type_uuid = model.uuid_fk_column(
        "inventory_adjustment_type.uuid", nullable=True
    )
    adjustment_type = orm.relationship(
        "InventoryAdjustmentType",
        doc="""
        Reference to the adjustment type record.
        """,
    )

    amount = sa.Column(
        sa.Numeric(precision=9, scale=4),
        nullable=False,
        doc="""
        Amount of the adjustment; may be positive or negative.
        """,
    )

    source = sa.Column(
        sa.String(length=100),
        nullable=True,
        doc="""
        Arbitrary string identifying the source of the adjustment, if applicable.
        """,
    )

    def __str__(self):
        return str(self.inventory or "")



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - data model
"""

from wuttjamaican.db.model import *

from .stores import Store
from .terminals import Terminal
from .tenders import Tender
from .taxes import Tax
from .employees import Employee
from .customers import Customer
from .departments import Department
from .products import (
    Product,
    ProductInventory,
    InventoryAdjustmentType,
    InventoryAdjustment,
)

from .batch.pos import POSBatch, POSBatchRow



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - feedback control
"""

import time

import flet as ft

from .keyboard import WuttaKeyboard
from wuttapos.terminal.util import show_snackbar


class WuttaFeedback(ft.Container):

    default_font_size = 20
    default_button_height = 60

    def __init__(self, config, page=None, *args, **kwargs):
        self.on_reset = kwargs.pop("on_reset", None)
        self.on_send = kwargs.pop("on_send", None)
        self.on_cancel = kwargs.pop("on_cancel", None)

        super().__init__(*args, **kwargs)

        self.config = config
        self.app = config.get_app()
        self.enum = self.app.enum

        # TODO: why must we save this aside from self.page ?
        # but sometimes self.page gets set to None, so we must..
        self.mypage = page

        self.content = ft.Text(
            "Feedback", size=self.default_font_size, weight=ft.FontWeight.BOLD
        )
        self.height = self.default_button_height
        self.width = self.default_button_height * 3
        self.on_click = self.initial_click
        self.alignment = ft.alignment.center
        self.border = ft.border.all(1, "black")
        self.border_radius = ft.border_radius.all(5)
        self.bgcolor = "blue"

    def informed_refresh(self, **kwargs):
        pass

    def reset(self, e=None):
        if self.on_reset:
            self.on_reset(e=e)

    def initial_click(self, e):

        self.message = ft.TextField(
            label="Message", multiline=True, min_lines=5, autofocus=True
        )

        self.dlg = ft.AlertDialog(
            modal=True,
            title=ft.Text("User Feedback"),
            content=ft.Container(
                content=ft.Column(
                    [
                        ft.Text(
                            "Questions, suggestions, comments, complaints, etc. "
                            "are welcome and may be submitted below. "
                        ),
                        ft.Divider(),
                        self.message,
                        ft.Divider(),
                        WuttaKeyboard(
                            self.config,
                            on_keypress=self.keypress,
                            on_long_backspace=self.long_backspace,
                        ),
                    ],
                    expand=True,
                ),
                height=800,
            ),
            actions=[
                ft.Row(
                    [
                        ft.Container(
                            content=ft.Text(
                                "Send Message",
                                size=self.default_font_size,
                                color="black",
                                weight=ft.FontWeight.BOLD,
                            ),
                            height=self.default_button_height,
                            width=self.default_button_height * 3,
                            alignment=ft.alignment.center,
                            bgcolor="blue",
                            border=ft.border.all(1, "black"),
                            border_radius=ft.border_radius.all(5),
                            on_click=self.send_feedback,
                        ),
                        ft.Container(
                            content=ft.Text(
                                "Cancel",
                                size=self.default_font_size,
                                weight=ft.FontWeight.BOLD,
                            ),
                            height=self.default_button_height,
                            width=self.default_button_height * 2.5,
                            alignment=ft.alignment.center,
                            border=ft.border.all(1, "black"),
                            border_radius=ft.border_radius.all(5),
                            on_click=self.cancel,
                        ),
                    ],
                    alignment=ft.MainAxisAlignment.CENTER,
                ),
            ],
        )

        # # TODO: leaving this for reference just in case..but hoping
        # the latest flet does not require this hack?

        # if self.mypage.dialog and self.mypage.dialog.open and self.mypage.dialog is not self.dlg:
        #     self.mypage.dialog.open = False
        #     self.mypage.update()
        #     # cf. https://github.com/flet-dev/flet/issues/1670
        #     time.sleep(0.1)

        # self.mypage.open(self.dlg)

        self.mypage.dialog = self.dlg
        self.dlg.open = True
        self.mypage.update()

    def keypress(self, key):
        if key == "⏎":
            self.message.value += "\n"
        elif key == "⌫":
            self.message.value = self.message.value[:-1]
        else:
            self.message.value += key

        self.message.focus()

        # TODO: why is keypress happening with no page?
        if self.page:
            self.update()

    def long_backspace(self):
        self.message.value = self.message.value[:-10]
        self.message.focus()
        self.update()

    def cancel(self, e):
        self.dlg.open = False
        self.mypage.update()

        if self.on_cancel:
            self.on_cancel(e)

    def send_feedback(self, e):
        if self.message.value:

            self.app.send_email(
                "pos_feedback",
                {
                    "user_name": self.mypage.session.get("user_display"),
                    "referrer": self.mypage.route,
                    "message": self.message.value,
                },
            )

            self.dlg.open = False
            show_snackbar(self.mypage, "MESSAGE WAS SENT", bgcolor="green")
            self.mypage.update()

            if self.on_send:
                self.on_send()

        else:
            self.message.focus()
            self.mypage.update()



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Common views
"""

from wuttaweb.views import common as base


class CommonView(base.CommonView):
    """
    Common views
    """

    def setup_enhance_admin_user(self, user):
        """ """
        model = self.app.model
        session = self.app.get_session(user)
        auth = self.app.get_auth_handler()

        site_admin = session.query(model.Role).filter_by(name="Site Admin").first()
        if site_admin:
            site_admin_perms = [
                "batch.pos.list",
                "batch.pos.view",
                "customers.list",
                "customers.create",
                "customers.view",
                "customers.edit",
                "departments.list",
                "departments.create",
                "departments.view",
                "departments.edit",
                "employees.list",
                "employees.create",
                "employees.view",
                "employees.edit",
                "inventory_adjustment_types.list",
                "inventory_adjustment_types.create",
                "inventory_adjustment_types.view",
                "inventory_adjustment_types.edit",
                "inventory_adjustments.list",
                "inventory_adjustments.create",
                "inventory_adjustments.view",
                "pos.test_error",
                "pos.ring_sales",
                "pos.override_price",
                "pos.del_customer",
                "pos.toggle_training",
                "pos.suspend",
                "pos.swap_customer",
                "pos.void_txn",
                "products.list",
                "products.create",
                "products.view",
                "products.edit",
                "stores.list",
                "stores.create",
                "stores.view",
                "stores.edit",
                "taxes.list",
                "taxes.create",
                "taxes.view",
                "taxes.edit",
                "tenders.list",
                "tenders.create",
                "tenders.view",
                "tenders.edit",
                "terminals.list",
                "terminals.create",
                "terminals.view",
                "terminals.edit",
            ]
            for perm in site_admin_perms:
                auth.grant_permission(site_admin, perm)


def defaults(config, **kwargs):
    local = globals()
    CommonView = kwargs.get("CommonView", local["CommonView"])
    base.defaults(config, **{"CommonView": CommonView})


def includeme(config):
    defaults(config)



"""add POS Batches

Revision ID: c2b1f0983c97
Revises: 7067ef686eb0
Create Date: 2026-01-02 13:11:37.703157

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "c2b1f0983c97"
down_revision: Union[str, None] = "7067ef686eb0"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # batch_pos
    op.create_table(
        "batch_pos",
        sa.Column("store_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("terminal_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("cashier_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("customer_uuid", wuttjamaican.db.util.UUID(), nullable=True),
        sa.Column("customer_is_member", sa.Boolean(), nullable=True),
        sa.Column("customer_is_employee", sa.Boolean(), nullable=True),
        sa.Column("sales_total", sa.Numeric(precision=9, scale=2), nullable=True),
        sa.Column("fs_total", sa.Numeric(precision=9, scale=2), nullable=True),
        sa.Column("tax_total", sa.Numeric(precision=9, scale=2), nullable=True),
        sa.Column("fs_tender_total", sa.Numeric(precision=9, scale=2), nullable=True),
        sa.Column("tender_total", sa.Numeric(precision=9, scale=2), nullable=True),
        sa.Column("void", sa.Boolean(), nullable=False),
        sa.Column("training_mode", sa.Boolean(), nullable=False),
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("description", sa.String(length=255), nullable=True),
        sa.Column("notes", sa.Text(), nullable=True),
        sa.Column("row_count", sa.Integer(), nullable=True),
        sa.Column("status_code", sa.Integer(), nullable=True),
        sa.Column("status_text", sa.String(length=255), nullable=True),
        sa.Column("created", sa.DateTime(), nullable=False),
        sa.Column("created_by_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("executed", sa.DateTime(), nullable=True),
        sa.Column("executed_by_uuid", wuttjamaican.db.util.UUID(), nullable=True),
        sa.ForeignKeyConstraint(
            ["cashier_uuid"],
            ["employee.uuid"],
            name=op.f("fk_batch_pos_cashier_uuid_employee"),
        ),
        sa.ForeignKeyConstraint(
            ["created_by_uuid"],
            ["user.uuid"],
            name=op.f("fk_batch_pos_created_by_uuid_user"),
        ),
        sa.ForeignKeyConstraint(
            ["customer_uuid"],
            ["customer.uuid"],
            name=op.f("fk_batch_pos_customer_uuid_customer"),
        ),
        sa.ForeignKeyConstraint(
            ["executed_by_uuid"],
            ["user.uuid"],
            name=op.f("fk_batch_pos_executed_by_uuid_user"),
        ),
        sa.ForeignKeyConstraint(
            ["store_uuid"], ["store.uuid"], name=op.f("fk_batch_pos_store_uuid_store")
        ),
        sa.ForeignKeyConstraint(
            ["terminal_uuid"],
            ["terminal.uuid"],
            name=op.f("fk_batch_pos_terminal_uuid_terminal"),
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_batch_pos")),
    )

    # batch_pos_row
    op.create_table(
        "batch_pos_row",
        sa.Column("modified_by_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("row_type", sa.String(length=20), nullable=False),
        sa.Column("item_entry", sa.String(length=20), nullable=True),
        sa.Column("product_uuid", wuttjamaican.db.util.UUID(), nullable=True),
        sa.Column("description", sa.String(length=100), nullable=True),
        sa.Column("foodstamp_eligible", sa.Boolean(), nullable=True),
        sa.Column("sold_by_weight", sa.Boolean(), nullable=True),
        sa.Column("quantity", sa.Numeric(precision=8, scale=2), nullable=True),
        sa.Column("cost", sa.Numeric(precision=8, scale=3), nullable=True),
        sa.Column("reg_price", sa.Numeric(precision=8, scale=3), nullable=True),
        sa.Column("cur_price", sa.Numeric(precision=8, scale=3), nullable=True),
        sa.Column("cur_price_type", sa.Integer(), nullable=True),
        sa.Column("cur_price_start", sa.DateTime(), nullable=True),
        sa.Column("cur_price_end", sa.DateTime(), nullable=True),
        sa.Column("txn_price", sa.Numeric(precision=8, scale=3), nullable=True),
        sa.Column("txn_price_adjusted", sa.Boolean(), nullable=True),
        sa.Column("sales_total", sa.Numeric(precision=9, scale=2), nullable=True),
        sa.Column("tax_code", sa.String(length=30), nullable=True),
        sa.Column("tender_total", sa.Numeric(precision=9, scale=2), nullable=True),
        sa.Column("void", sa.Boolean(), nullable=False),
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("batch_uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("sequence", sa.Integer(), nullable=False),
        sa.Column("status_code", sa.Integer(), nullable=True),
        sa.Column("status_text", sa.String(length=255), nullable=True),
        sa.Column("modified", sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(
            ["batch_uuid"],
            ["batch_pos.uuid"],
            name=op.f("fk_batch_pos_row_batch_uuid_batch_pos"),
        ),
        sa.ForeignKeyConstraint(
            ["modified_by_uuid"],
            ["user.uuid"],
            name=op.f("fk_batch_pos_row_modified_by_uuid_user"),
        ),
        sa.ForeignKeyConstraint(
            ["product_uuid"],
            ["product.uuid"],
            name=op.f("fk_batch_pos_row_product_uuid_product"),
        ),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_batch_pos_row")),
    )


def downgrade() -> None:

    # batch_pos*
    op.drop_table("batch_pos_row")
    op.drop_table("batch_pos")



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Model for Departments
"""

import sqlalchemy as sa
from sqlalchemy import orm

from wuttjamaican.db import model


class Department(model.Base):
    """
    Represents an organizational department, for products and/or personnel.
    """

    __tablename__ = "department"
    __versioned__ = {}
    __wutta_hint__ = {
        "model_title": "Department",
        "model_title_plural": "Departments",
    }

    uuid = model.uuid_column()

    department_id = sa.Column(
        sa.String(length=20),
        nullable=False,
        doc="""
        Unique identifier for the department.
        """,
    )

    name = sa.Column(
        sa.String(length=100),
        nullable=False,
        doc="""
        Name of the department.
        """,
    )

    for_products = sa.Column(
        sa.Boolean(),
        nullable=False,
        doc="""
        Indicates the department exists to organize products.
        """,
    )

    for_personnel = sa.Column(
        sa.Boolean(),
        nullable=False,
        doc="""
        Indicates the department exists to organize personnel.
        """,
    )

    exempt_from_gross_sales = sa.Column(
        sa.Boolean(),
        nullable=True,
        doc="""
        Indicates products in this department do not count toward gross sales.
        """,
    )

    def __str__(self):
        return self.name or ""



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
POS batch handler
"""

# import decimal

import sqlalchemy as sa
from sqlalchemy import orm

from wuttjamaican.batch import BatchHandler

from wuttapos.db.model import POSBatch


class POSBatchHandler(BatchHandler):
    """
    Handler for POS batches
    """

    model_class = POSBatch

    def get_store_id(self, require=False):
        """
        Returns the ID string for store to which local terminal belongs.
        """
        return self.config.get("wuttapos.store_id", require=require)

    def get_terminal_id(self, require=False):
        """
        Returns the ID string for current POS terminal.
        """
        return self.config.get("wuttapos.terminal_id", require=require)

    def init_batch(self, batch, session=None, **kwargs):
        if not session:
            raise ValueError(
                f"must provide session to {self.__class__.__name__}init_batch()"
            )

        model = self.app.model

        batch.store = (
            session.query(model.Store)
            .filter_by(store_id=self.get_store_id(require=True))
            .one()
        )

        batch.terminal = (
            session.query(model.Terminal)
            .filter_by(terminal_id=self.get_terminal_id(require=True))
            .one()
        )

        batch.status_code = batch.STATUS_OK

    # TODO: should also filter this by terminal
    def get_current_batch(
        self, user, training_mode=False, create=True, return_created=False
    ):
        """
        Get the "current" POS batch for the given user, creating it as
        needed.

        :param user: Reference to active user (cashier) for the batch
           / POS transaction.

        :param training_mode: Flag indicating whether the transaction
           should be in training mode.  The lookup will be restricted
           according to the value of this flag.  If a new batch is
           created, it will be assigned this flag value.

        :param create: Whether a new batch should be created, if no
           current batch is found.

        :param return_created: Indicates the return value should
           include both the batch *and* a flag indicating whether the
           batch was newly-created.
        """
        if not user:
            raise ValueError("must specify a user")

        created = False
        model = self.app.model
        session = self.app.get_session(user)
        employee = self.app.get_employee(user)

        # TODO: can we assume cashier (employee) is always set /
        # accurate?  if so maybe stop filtering on created_by?
        cashier_criteria = sa.and_(
            model.POSBatch.cashier == None, model.POSBatch.created_by == user
        )
        if employee:
            cashier_criteria = sa.or_(
                model.POSBatch.cashier == employee, cashier_criteria
            )

        try:
            batch = (
                session.query(model.POSBatch)
                .filter(cashier_criteria)
                .filter(model.POSBatch.status_code == model.POSBatch.STATUS_OK)
                .filter(model.POSBatch.executed == None)
                .filter(model.POSBatch.training_mode == training_mode)
                .one()
            )

        except orm.exc.NoResultFound:
            if not create:
                if return_created:
                    return None, False
                return None
            batch = self.make_batch(
                session, created_by=user, training_mode=training_mode, cashier=employee
            )
            session.add(batch)
            session.flush()
            created = True

        if return_created:
            return batch, created

        return batch

    def get_screen_txn_display(self, batch):
        """
        Should return the text to be used for displaying transaction
        identifier within the header of POS screen.
        """
        return batch.id_str

    def get_screen_cust_display(self, batch=None, customer=None):
        """
        Should return the text to be used for displaying customer
        identifier / name etc. within the header of POS screen.
        """
        if not customer and batch:
            customer = batch.customer
        if customer:
            return customer.customer_id
        return None

    # TODO: this should account for shoppers somehow too
    def set_customer(self, batch, customer, user=None, **kwargs):
        """
        Assign the customer account for POS transaction.
        """
        enum = self.app.enum

        if customer and batch.customer:
            row_type = enum.POS_ROW_TYPE_SWAP_CUSTOMER
        elif customer:
            row_type = enum.POS_ROW_TYPE_SET_CUSTOMER
        else:
            if not batch.customer:
                return
            row_type = enum.POS_ROW_TYPE_DEL_CUSTOMER

        batch.customer = customer
        if customer:
            # member = self.app.get_member(customer)
            # batch.customer_is_member = bool(member and member.active)
            # batch.customer_is_member = False
            batch.customer_is_member = None
            employee = self.app.get_employee(customer)
            batch.customer_is_employee = bool(employee and employee.active)
        else:
            batch.customer_is_member = None
            batch.customer_is_employee = None

        row = self.make_row()
        row.modified_by = user
        row.row_type = row_type
        if customer:
            row.item_entry = customer.customer_id
        row.description = (
            f"{customer.customer_id} - {customer.name}"
            if customer
            else "REMOVE CUSTOMER"
        )
        self.add_row(batch, row)

    def process_entry(self, batch, entry, quantity=1, user=None, **kwargs):
        """
        Process an "entry" value direct from POS.  Most typically,
        this is effectively "ringing up an item" and hence we add a
        row to the batch and return the row.
        """
        session = self.app.get_session(batch)
        model = self.app.model
        enum = self.app.enum

        if isinstance(entry, model.Product):
            product = entry
            entry = product.uuid
        else:
            # TODO: maybe should only search by product key (or GPC?) etc.
            product = self.app.get_products_handler().locate_product_for_entry(
                session, entry
            )

        # TODO: if product not found, should auto-record a badscan
        # entry?  maybe only if config says so, e.g. might be nice to
        # only record badscan if entry truly came from scanner device,
        # in which case only the caller would know that
        if not product:
            return None

        # if product.not_for_sale:
        #     key = self.app.get_products_handler().render_product_key(product)
        #     raise ValueError(f"product is not for sale: {key}")

        # product located, so add item row
        row = self.make_row()
        row.modified_by = user
        row.item_entry = kwargs.get("item_entry", entry)
        row.product = product
        # row.brand_name = product.brand.name if product.brand else None
        row.description = product.description
        # row.size = product.size
        # dept = product.department
        # if dept:
        #     row.department_number = dept.number
        #     row.department_name = dept.name
        # row.department = product.department
        # subdept = product.subdepartment
        # if subdept:
        #     row.subdepartment_number = subdept.number
        #     row.subdepartment_name = subdept.name
        # row.foodstamp_eligible = product.food_stampable
        row.sold_by_weight = product.sold_by_weight
        row.quantity = quantity

        # regprice = product.regular_price
        # if regprice:
        #     row.reg_price = regprice.price

        row.reg_price = product.unit_price_reg

        # curprice = product.current_price
        # if curprice:
        #     row.cur_price = curprice.price
        #     row.cur_price_type = curprice.type
        #     row.cur_price_start = curprice.starts
        #     row.cur_price_end = curprice.ends

        # row.txn_price = row.cur_price or row.reg_price
        row.txn_price = row.reg_price

        if row.txn_price:
            row.sales_total = row.txn_price * row.quantity
            batch.sales_total = (batch.sales_total or 0) + row.sales_total
            # if row.foodstamp_eligible:
            #     batch.fs_total = (batch.fs_total or 0) + row.sales_total

        # tax = product.tax
        # if tax:
        #     row.tax_code = tax.code

        if row.txn_price:
            row.row_type = enum.POS_ROW_TYPE_SELL
            # if tax:
            #     self.update_tax(batch, row, tax)
        else:
            row.row_type = enum.POS_ROW_TYPE_BADPRICE

        self.add_row(batch, row)
        session.flush()
        return row

    # def update_tax(self, batch, row, tax=None, tax_code=None, **kwargs):
    #     """
    #     Update the tax totals for the batch, basd on given row.
    #     """
    #     if not tax and not tax_code:
    #         raise ValueError("must specify either tax or tax_code")

    #     session = self.app.get_session(batch)
    #     if not tax:
    #         tax = self.get_tax(session, tax_code)

    #     btax = batch.taxes.get(tax.code)
    #     if not btax:
    #         btax = self.model.POSBatchTax()
    #         btax.tax = tax
    #         btax.tax_code = tax.code
    #         btax.tax_rate = tax.rate
    #         session.add(btax)
    #         btax.batch = batch
    #         session.flush()

    #     # calculate relevant sales
    #     rows = [r for r in batch.active_rows() if r.tax_code == tax.code and not r.void]
    #     sales = sum([r.sales_total for r in rows])
    #     # nb. must add row separately if not yet in batch
    #     if not row.batch and not row.batch_uuid:
    #         sales += row.sales_total

    #     # total for this tax
    #     before = btax.tax_total or 0
    #     btax.tax_total = (sales * (tax.rate / 100)).quantize(decimal.Decimal("0.02"))
    #     batch.tax_total = (batch.tax_total or 0) - before + btax.tax_total

    # def record_badscan(self, batch, entry, quantity=1, user=None, **kwargs):
    #     """
    #     Add a row to the batch which represents a "bad scan" at POS.
    #     """
    #     row = self.make_row()
    #     row.user = user
    #     row.row_type = self.enum.POS_ROW_TYPE_BADSCAN
    #     row.item_entry = entry
    #     row.description = "BADSCAN"
    #     row.quantity = quantity
    #     self.add_row(batch, row)
    #     return row

    def add_open_ring(self, batch, department, price, quantity=1, user=None, **kwargs):
        """
        Adds an "open ring" row to the batch.
        """
        model = self.app.model
        enum = self.app.enum
        session = self.app.get_session(batch)

        if not isinstance(department, model.Department):
            department = session.get(model.Department, department)
            if not department:
                raise ValueError("must specify valid department")

        # add row for open ring
        row = self.make_row()
        row.row_type = enum.POS_ROW_TYPE_OPEN_RING
        row.modified_by = user
        row.item_entry = department.department_id
        row.description = f"OPEN RING: {department.name}"
        # row.department_number = department.number
        # row.department_name = department.name
        # row.department = department
        # row.foodstamp_eligible = department.food_stampable
        row.quantity = quantity

        row.txn_price = price
        if row.txn_price:
            row.sales_total = row.txn_price * row.quantity
            batch.sales_total = (batch.sales_total or 0) + row.sales_total
            # if row.foodstamp_eligible:
            #     batch.fs_total = (batch.fs_total or 0) + row.sales_total

        # tax = department.tax
        # if tax:
        #     row.tax_code = tax.code
        #     self.update_tax(batch, row, tax)

        self.add_row(batch, row)
        session.flush()
        return row

    # def get_tax(self, session, code, **kwargs):
    #     """
    #     Return the tax record corresponding to the given code.

    #     :param session: Current DB session.

    #     :param code: Tax code to fetch.
    #     """
    #     model = self.model
    #     return session.query(model.Tax).filter(model.Tax.code == code).one()

    # def get_tender(self, session, key, **kwargs):
    #     """
    #     Return the tender record corresponding to the given key.

    #     :param session: Current DB session.

    #     :param key: Either a tender UUID, or "true" tender code (i.e.
    #        :attr:`rattail.db.model.sales.Tender.code` value) or a
    #        "pseudo-code" for common tenders (e.g. ``'cash'``).
    #     """
    #     model = self.model

    #     # Tender.uuid match?
    #     tender = session.get(model.Tender, key)
    #     if tender:
    #         return tender

    #     # Tender.code match?
    #     try:
    #         return session.query(model.Tender).filter(model.Tender.code == key).one()
    #     except orm.exc.NoResultFound:
    #         pass

    #     # try settings, if value then recurse
    #     # TODO: not sure why get_vendor() only checks settings?
    #     # for now am assuming we should also check config file
    #     # key = self.app.get_setting(session, f'rattail.tender.{key}')
    #     key = self.config.get("rattail", f"tender.{key}")
    #     if key:
    #         return self.get_tender(session, key, **kwargs)

    def refresh_row(self, row):
        # TODO (?)
        row.status_code = row.STATUS_OK
        row.status_text = None

    # # TODO: this subclass used to override the base method here, but for
    # # the moment we implement the full method below, sans customization.
    # def clone(self, oldbatch, created_by, **kwargs):
    #     newbatch = super().clone(oldbatch, created_by, **kwargs)
    #     session = self.app.get_session(oldbatch)
    #     model = self.app.model

    #     # tax_mapper = sa.inspect(model.POSBatchTax)
    #     # for oldtax in oldbatch.taxes.values():
    #     #     newtax = model.POSBatchTax()
    #     #     for key in tax_mapper.columns.keys():
    #     #         if key not in ("uuid", "batch_uuid"):
    #     #             setattr(newtax, key, getattr(oldtax, key))
    #     #     session.add(newtax)
    #     #     newtax.batch = newbatch

    #     return newbatch

    # TODO: this used to be in the BatchHandler base class, and i
    # guess it probably still should be, once it's vetted.  also
    # see the note above for custom clone() method.
    def clone(self, oldbatch, created_by, progress=None, **kwargs):
        """
        Clone the given batch as a new batch, and return the new batch.
        """
        session = self.app.get_session(oldbatch)

        # self.setup_clone(oldbatch, progress=progress)
        batch_class = self.model_class
        batch_mapper = sa.inspect(batch_class)

        newbatch = batch_class()
        newbatch.id = self.consume_batch_id(session)
        newbatch.created_by = created_by
        newbatch.row_count = 0
        for name in batch_mapper.columns.keys():
            if name not in (
                "uuid",
                "id",
                "created",
                "created_by_uuid",
                "row_count",
                "executed",
                "executed_by_uuid",
            ):
                setattr(newbatch, name, getattr(oldbatch, name))

        session.add(newbatch)
        session.flush()

        row_class = newbatch.__row_class__
        row_mapper = sa.inspect(row_class)

        def clone_row(oldrow, i):
            newrow = self.clone_row(oldrow)
            self.add_row(newbatch, newrow)

        self.app.progress_loop(
            clone_row,
            self.get_clonable_rows(oldbatch),
            progress,
            message="Cloning data rows for new batch",
        )

        self.refresh_batch_status(newbatch)
        # self.teardown_clone(newbatch, progress=progress)
        return newbatch

    # # TODO: this used to be defined by the base method as shown here
    # def get_clonable_rows(self, batch, **kwargs):
    #     return batch.data_rows

    # TODO: for now we are providing the only implementation; see above
    def get_clonable_rows(self, batch, **kwargs):
        enum = self.app.enum
        # TODO: row types..ugh
        return [row for row in batch.rows if row.row_type != enum.POS_ROW_TYPE_TENDER]

    # # TODO: this used to be defined by the base method as shown here (i think)
    def clone_row(self, oldrow):
        row_class = self.model_class.__row_class__
        row_mapper = sa.inspect(row_class)
        newrow = row_class()
        for name in row_mapper.columns.keys():
            if name not in ("uuid", "batch_uuid", "sequence"):
                setattr(newrow, name, getattr(oldrow, name))
        return newrow

    def override_price(self, row, user, txn_price, **kwargs):
        """
        Override the transaction price for the given batch row.
        """
        enum = self.app.enum
        batch = row.batch

        # update price for given row
        orig_row = row
        orig_txn_price = orig_row.txn_price
        orig_sales_total = orig_row.sales_total
        orig_row.txn_price = txn_price
        orig_row.txn_price_adjusted = True
        orig_row.sales_total = orig_row.quantity * orig_row.txn_price

        # adjust totals
        batch.sales_total = (
            (batch.sales_total or 0) - orig_sales_total + orig_row.sales_total
        )
        # if orig_row.foodstamp_eligible:
        #     batch.fs_total = (
        #         (batch.fs_total or 0) - orig_sales_total + orig_row.sales_total
        #     )
        # if orig_row.tax_code:
        #     self.update_tax(batch, orig_row, tax_code=orig_row.tax_code)

        # add another row indicating who/when
        row = self.make_row()
        row.modified_by = user
        row.row_type = enum.POS_ROW_TYPE_ADJUST_PRICE
        row.item_entry = orig_row.item_entry
        row.txn_price = txn_price
        row.description = (
            f"ROW {orig_row.sequence} PRICE ADJUST "
            f"FROM {self.app.render_currency(orig_txn_price)}"
        )
        self.add_row(batch, row)
        return row

    def void_row(self, row, user, **kwargs):
        """
        Apply "void" status to the given batch row.
        """
        enum = self.app.enum
        batch = row.batch

        # mark given row as void
        orig_row = row
        orig_row.void = True

        # adjust batch totals
        if orig_row.sales_total:
            batch.sales_total = (batch.sales_total or 0) - orig_row.sales_total
            # if orig_row.foodstamp_eligible:
            #     batch.fs_total = (batch.fs_total or 0) - orig_row.sales_total
        # if orig_row.tax_code:
        #     self.update_tax(batch, orig_row, tax_code=orig_row.tax_code)

        # add another row indicating who/when
        row = self.make_row()
        row.modified_by = user
        row.row_type = enum.POS_ROW_TYPE_VOID_LINE
        row.item_entry = orig_row.item_entry
        row.description = f"VOID ROW {orig_row.sequence}"
        self.add_row(batch, row)
        return row

    def suspend_transaction(self, batch, user, **kwargs):
        """
        Suspend transaction for the given POS batch.
        """
        enum = self.app.enum

        # add another row indicating who/when
        row = self.make_row()
        row.modified_by = user
        row.row_type = enum.POS_ROW_TYPE_SUSPEND
        row.description = "SUSPEND TXN"
        self.add_row(batch, row)

        # TODO: should do something different if we have a server
        # engine, i.e. central location for suspended txns

        # mark batch as suspended (but not executed)
        batch.status_code = batch.STATUS_SUSPENDED

    def resume_transaction(self, batch, user, **kwargs):
        """
        Resume transaction for the given POS batch.  By default this
        always creates a *new* batch.
        """
        enum = self.app.enum

        # TODO: should do something different if we have a server
        # engine, i.e. central location for suspended txns

        newbatch = self.clone(batch, user)
        newbatch.cashier = self.app.get_employee(user)
        newbatch.status_code = newbatch.STATUS_OK

        # add another row indicating who/when
        row = self.make_row()
        row.modified_by = user
        row.row_type = enum.POS_ROW_TYPE_RESUME
        row.description = "RESUME TXN"
        self.add_row(newbatch, row)

        # mark original batch as executed
        batch.executed = self.app.make_utc()
        batch.executed_by = user

        return newbatch

    def void_batch(self, batch, user, **kwargs):
        """
        Void the given POS batch.
        """
        enum = self.app.enum

        # add another row indicating who/when
        row = self.make_row()
        row.modified_by = user
        row.row_type = enum.POS_ROW_TYPE_VOID_TXN
        row.description = "VOID TXN"
        self.add_row(batch, row)

        # void/execute batch
        batch.void = True
        batch.executed = self.app.make_utc()
        batch.executed_by = user

    # def apply_tender(self, batch, user, tender, amount, **kwargs):
    #     """
    #     Apply the given tender amount to the batch.

    #     :param tender: Reference to a
    #        :class:`~rattail.db.model.sales.Tender` or similar object, or dict
    #        with similar keys, or can be just a tender code.

    #     :param amount: Amount to apply.  Note, this usually should be
    #        a *negative* number.

    #     :returns: List of rows which were added to the batch.
    #     """
    #     session = self.app.get_session(batch)
    #     model = self.model

    #     tender_info = tender
    #     tender = None
    #     item_entry = None
    #     description = None

    #     if isinstance(tender_info, model.Tender):
    #         tender = tender_info
    #     else:
    #         if isinstance(tender_info, str):
    #             tender_code = tender_info
    #         else:
    #             tender_code = tender_info["code"]
    #         item_entry = tender_code
    #         description = f"TENDER '{tender_code}'"
    #         tender = self.get_tender(session, tender_code)

    #     if tender:
    #         item_entry = tender.code
    #         description = tender.name

    #         if tender.disabled:
    #             # TODO: behavior here should be configurable, probably
    #             # needs a dedicated email etc. ..or maybe just ignore
    #             log.error(
    #                 "disabled tender '%s' being applied to POS batch: %s", tender, batch
    #             )

    #     rows = []

    #     # add row for tender
    #     row = self.make_row()
    #     row.user = user
    #     row.row_type = self.enum.POS_ROW_TYPE_TENDER
    #     row.item_entry = item_entry
    #     row.description = description
    #     row.tender_total = amount
    #     row.tender = tender
    #     batch.tender_total = (batch.tender_total or 0) + row.tender_total
    #     if tender and tender.is_foodstamp:
    #         batch.fs_tender_total = (batch.fs_tender_total or 0) + row.tender_total
    #     self.add_row(batch, row)
    #     rows.append(row)

    #     # nothing more to do for now, if balance remains
    #     balance = batch.get_balance()
    #     if balance > 0:
    #         return rows

    #     # next we'll give change back
    #     # nb. if balance is 0, then change due is 0, but we always
    #     # include the change due line item even so..

    #     # ..but some tenders do not allow cash back
    #     if balance < 0:
    #         if hasattr(tender, "is_cash") and not tender.is_cash:
    #             if not tender.allow_cash_back:
    #                 raise ValueError(
    #                     f"tender '{tender.name}' does not allow "
    #                     f" cash back: ${-balance:0.2f}"
    #                 )

    #     # TODO: maybe should always give change as 'cash'
    #     if tender and tender.is_cash:
    #         cash = tender
    #     else:
    #         cash = self.get_tender(session, "cash")

    #     row = self.make_row()
    #     row.user = user
    #     row.row_type = self.enum.POS_ROW_TYPE_CHANGE_BACK
    #     row.item_entry = item_entry
    #     row.description = "CHANGE DUE"
    #     row.tender_total = -balance
    #     row.tender = cash
    #     batch.tender_total = (batch.tender_total or 0) + row.tender_total
    #     self.add_row(batch, row)
    #     rows.append(row)

    #     # all paid up, so finalize
    #     session.flush()
    #     assert batch.get_balance() == 0
    #     self.do_execute(batch, user, **kwargs)
    #     return rows

    # def execute(self, batch, progress=None, **kwargs):
    #     # TODO
    #     return True



# -*- coding: utf-8; -*-
"""
Master view for Taxes
"""

from wuttapos.db.model.taxes import Tax

from wuttaweb.views import MasterView


class TaxView(MasterView):
    """
    Master view for Taxes
    """

    model_class = Tax
    model_title = "Tax"
    model_title_plural = "Taxes"

    route_prefix = "taxes"
    url_prefix = "/taxes"

    creatable = True
    editable = True
    deletable = True

    labels = {
        "tax_id": "Tax ID",
    }

    grid_columns = [
        "tax_id",
        "name",
        "rate",
    ]

    form_fields = [
        "tax_id",
        "name",
        "rate",
    ]


def defaults(config, **kwargs):
    base = globals()

    TaxView = kwargs.get("TaxView", base["TaxView"])
    TaxView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
"""
Model definition for Employees
"""

import sqlalchemy as sa
from sqlalchemy import orm

from wuttjamaican.db import model


class Employee(model.Base):
    """
    Represents a current or former employee.
    """

    __tablename__ = "employee"
    __versioned__ = {}
    __wutta_hint__ = {
        "model_title": "Employee",
        "model_title_plural": "Employees",
    }

    uuid = model.uuid_column()

    person_uuid = model.uuid_fk_column("person.uuid", nullable=True)
    person = orm.relationship(
        "Person",
        doc="""
        Reference to the person who is/was the employee.
        """,
        backref=orm.backref(
            "employee",
            uselist=False,
            doc="""
            Reference to the employee record for the person, if applicable.
            """,
        ),
    )

    name = sa.Column(
        sa.String(length=100),
        nullable=False,
        doc="""
        Internal name for the employee.
        """,
    )

    public_name = sa.Column(
        sa.String(length=100),
        nullable=True,
        doc="""
        Name of the employee, for display to the public (if different).
        """,
    )

    active = sa.Column(
        sa.Boolean(),
        nullable=False,
        default=True,
        doc="""
        Indicates the employee is currently active.
        """,
    )

    def __str__(self):
        return self.name or ""



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Master view for POS Batches
"""

from wuttaweb.views.batch import BatchMasterView

from wuttapos.db.model import POSBatch
from wuttapos.server.forms.schema import StoreRef, TerminalRef, EmployeeRef, CustomerRef


class POSBatchView(BatchMasterView):
    """
    Master view for POS Batches
    """

    model_class = POSBatch

    route_prefix = "batch.pos"
    url_prefix = "/batch/pos"

    creatable = False
    editable = False
    # nb. allow delete for now, at least is useful in dev?
    deletable = True
    executable = False

    labels = {
        "terminal_id": "Terminal ID",
    }

    grid_columns = [
        "id",
        "created",
        "store",
        "terminal",
        "cashier",
        "customer",
        "row_count",
        "sales_total",
        "void",
        "training_mode",
        "status_code",
        "executed",
    ]

    form_fields = [
        "id",
        "terminal",
        "cashier",
        "customer",
        "customer_is_member",
        "customer_is_employee",
        "params",
        "row_count",
        "sales_total",
        # 'taxes',
        "tender_total",
        "fs_tender_total",
        "balance",
        "void",
        "training_mode",
        "status_code",
        "created",
        "created_by",
        "executed",
        "executed_by",
    ]

    filter_defaults = {
        "executed": {"active": True, "verb": "is_null"},
    }

    row_grid_columns = [
        "sequence",
        "row_type",
        "item_entry",
        "description",
        "product",
        "reg_price",
        "txn_price",
        "quantity",
        "sales_total",
        "tender_total",
        "tax_code",
        "modified_by",
    ]

    def get_batch_handler(self):
        """
        Must return the :term:`batch handler` for use with this view.

        There is no default logic; subclass must override.
        """
        spec = "wuttapos.batch.pos:POSBatchHandler"
        factory = self.app.load_object(spec)
        return factory(self.config)

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # links
        g.set_link("terminal")
        g.set_link("cashier")
        g.set_link("customer")

    def grid_row_class(self, batch, data, i):
        if batch.training_mode:
            return "has-background-warning"
        if batch.void:
            return "has-background-danger-light"
        return None

    def configure_form(self, form):
        f = form
        super().configure_form(f)

        # store
        f.set_node("store", StoreRef(self.request))

        # terminal
        f.set_node("terminal", TerminalRef(self.request))

        # cashier
        f.set_node("cashier", EmployeeRef(self.request))

        # customer
        f.set_node("customer", CustomerRef(self.request))


def defaults(config, **kwargs):
    base = globals()

    POSBatchView = kwargs.get("POSBatchView", base["POSBatchView"])
    POSBatchView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS app
"""

import logging
import socket
import sys
import threading
from collections import OrderedDict
from traceback import format_exception

from wuttjamaican.conf import make_config
from wuttjamaican.util import resource_path

import flet as ft

import wuttapos
from wuttapos.terminal.controls.buttons import make_button
from wuttapos.terminal.util import get_pos_batch_handler


log = logging.getLogger(__name__)


def main(page: ft.Page):
    config = make_config()
    app = config.get_app()
    model = app.model
    handler = get_pos_batch_handler(config)

    # nb. as of python 3.10 the original hook is accessible, we needn't save the ref
    # cf. https://docs.python.org/3/library/threading.html#threading.__excepthook__
    orig_thread_hook = threading.excepthook

    hostname = socket.gethostname()
    email_context = OrderedDict(
        [
            ("hostname", hostname),
            ("ipaddress", socket.gethostbyname(hostname)),
            ("terminal", handler.get_terminal_id() or "??"),
        ]
    )

    def handle_error(exc_type, exc_value, exc_traceback):

        log.exception("unhandled error in POS")

        # nb. ignore this particular error; it is benign
        if exc_type is RuntimeError and str(exc_value) == "Event loop is closed":
            log.debug("ignoring error for closed event loop", exc_info=True)
            return

        extra_context = OrderedDict(email_context)
        traceback = "".join(
            format_exception(exc_type, exc_value, exc_traceback)
        ).strip()

        try:
            uuid = page.session.get("user_uuid")
            if uuid:
                session = app.make_session()
                user = session.get(model.User, uuid)
                extra_context["username"] = user.username
                # TODO
                # batch = handler.get_current_batch(user, create=False)
                # if batch:
                #     extra_context['batchid'] = batch.id_str
                session.close()
            else:
                extra_context["username"] = "n/a"

            app.send_email(
                "uncaught_exception",
                {
                    "extra_context": extra_context,
                    "error": app.render_error(exc_value),
                    "traceback": traceback,
                },
            )

        except:
            log.exception("failed to send error email")

        try:

            def close_bs(e):
                bs.open = False
                bs.update()

            bs = ft.BottomSheet(
                ft.Container(
                    ft.Column(
                        [
                            ft.Text(
                                "Unexpected Error", size=24, weight=ft.FontWeight.BOLD
                            ),
                            ft.Divider(),
                            ft.Text(
                                "Please be advised, something unexpected has gone wrong.\n"
                                "The state of your transaction may be questionable.\n\n"
                                "If possible you should consult the IT administrator.\n"
                                "(They may have already received an email about this.)",
                                size=20,
                            ),
                            ft.Container(
                                content=make_button(
                                    "Dismiss", on_click=close_bs, height=80, width=120
                                ),
                                alignment=ft.alignment.center,
                                expand=1,
                            ),
                        ],
                    ),
                    bgcolor="yellow",
                    padding=20,
                ),
                open=True,
                dismissible=False,
            )

            page.overlay.append(bs)
            page.update()

        except:
            log.exception("failed to show error bottomsheet")

    def sys_exc_hook(exc_type, exc_value, exc_traceback):
        handle_error(exc_type, exc_value, exc_traceback)
        sys.__excepthook__(exc_type, exc_value, exc_traceback)

    def thread_exc_hook(args):
        handle_error(args.exc_type, args.exc_value, args.exc_traceback)
        # nb. as of python 3.10 could just call threading.__excepthook__ instead
        # cf. https://docs.python.org/3/library/threading.html#threading.__excepthook__
        orig_thread_hook(args)

    # custom exception hook for main process
    sys.excepthook = sys_exc_hook

    # custom exception hook for threads (requires python 3.8)
    # cf. https://docs.python.org/3/library/threading.html#threading.excepthook
    v = sys.version_info
    if v.major >= 3 and v.minor >= 8:
        threading.excepthook = thread_exc_hook

    page.title = f"WuttaPOS v{wuttapos.__version__}"
    if hasattr(page, "window"):
        page.window.full_screen = True
    else:
        page.window_full_screen = True

    # global defaults for button/text styles etc.
    page.data = {
        "default_button_height_pos": 100,
        "default_button_height_dlg": 80,
    }

    def clean_exit():
        # TODO: this doesn't do anything interesting now, but it could
        if hasattr(page, "window"):
            page.window.destroy()
        else:
            page.window_destroy()

    def keyboard(e):
        # exit on ctrl+Q
        if e.ctrl and e.key == "Q":
            if not e.shift and not e.alt and not e.meta:
                clean_exit()

    page.on_keyboard_event = keyboard

    def window_event(e):
        if e.data == "close":
            clean_exit()

    # cf. https://flet.dev/docs/controls/page/#window_destroy
    if hasattr(page, "window"):
        page.window.prevent_close = True
        page.window.on_event = window_event
    else:
        page.window_prevent_close = True
        page.window_on_event = window_event

    # TODO: probably these should be auto-loaded from spec
    from wuttapos.terminal.views.pos import POSView
    from wuttapos.terminal.views.login import LoginView

    # cf .https://flet.dev/docs/guides/python/navigation-and-routing#building-views-on-route-change

    def route_change(e):
        page.views.clear()

        redirect = None
        user_uuid = page.session.get("user_uuid")
        if page.route == "/login" and user_uuid:
            redirect = "/pos"
            other = "/pos"
        elif page.route == "/pos" and not user_uuid:
            redirect = "/login"
            other = "/login"
        else:
            redirect = "/pos" if user_uuid else "/login"

        if redirect and page.route != redirect:
            page.go(redirect)
            return

        if page.route == "/pos":
            page.views.append(POSView(config, "/pos"))

        elif page.route == "/login":
            page.views.append(LoginView(config, "/login"))

        if hasattr(page, "window"):
            page.window.full_screen = True
        else:
            page.window_full_screen = True

        page.update()

    # TODO: this was in example docs but not sure what it's for?
    # def view_pop(view):
    #     page.views.pop()
    #     top_view = page.views[-1]
    #     page.go(top_view.route)

    page.on_route_change = route_change
    # page.on_view_pop = view_pop

    # TODO: this may be too hacky but is useful for now/dev
    if not config.production():

        training = page.client_storage.get("training")
        page.session.set("training", training)

        user = None
        uuid = page.client_storage.get("user_uuid")
        if uuid:
            session = app.make_session()
            user = session.get(model.User, uuid)
            if user:
                page.session.set("user_uuid", user.uuid.hex)
                page.session.set("user_display", str(user))

                if batch := handler.get_current_batch(user, create=False):
                    page.session.set("txn_display", batch.id_str)
                    if batch.customer:
                        page.session.set("cust_uuid", batch.customer.uuid)
                        page.session.set(
                            "cust_display", handler.get_screen_cust_display(txn=txn)
                        )

                session.close()
                page.go("/pos")
                return

            session.close()

    page.go("/login")


# TODO: can we inject config to the main() via ft.app() kwargs somehow?
# pretty sure the `wuttapos open` command is trying to anyway..
def run_app(config=None):
    ft.app(target=main, assets_dir=resource_path("wuttapos.terminal:assets"))


if __name__ == "__main__":
    run_app()



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS server menu
"""

from wuttaweb import menus as base


class WuttaPosMenuHandler(base.MenuHandler):
    """
    WuttaPOS menu handler
    """

    def make_menus(self, request, **kwargs):
        return [
            self.make_people_menu(request),
            self.make_products_menu(request),
            self.make_batches_menu(request),
            self.make_admin_menu(request),
        ]

    def make_people_menu(self, request):
        return {
            "title": "People",
            "type": "menu",
            "items": [
                {
                    "title": "Customers",
                    "route": "customers",
                    "perm": "customers.list",
                },
                {
                    "title": "Employees",
                    "route": "employees",
                    "perm": "employees.list",
                },
                {
                    "title": "All People",
                    "route": "people",
                    "perm": "people.list",
                },
            ],
        }

    def make_products_menu(self, request):
        return {
            "title": "Products",
            "type": "menu",
            "items": [
                {
                    "title": "Products",
                    "route": "products",
                    "perm": "products.list",
                },
                {"type": "sep"},
                {
                    "title": "Departments",
                    "route": "departments",
                    "perm": "departments.list",
                },
                {"type": "sep"},
                {
                    "title": "Inventory Adjustments",
                    "route": "inventory_adjustments",
                    "perm": "inventory_adjustments.list",
                },
                {
                    "title": "New Inventory Adjustment",
                    "route": "inventory_adjustments.create",
                    "perm": "inventory_adjustments.create",
                },
                {
                    "title": "Inventory Adjustment Types",
                    "route": "inventory_adjustment_types",
                    "perm": "inventory_adjustment_types.list",
                },
                # {
                #     "title": "Vendors",
                #     "route": "vendors",
                #     "perm": "vendors.list",
                # },
            ],
        }

    def make_batches_menu(self, request):
        return {
            "title": "Batches",
            "type": "menu",
            "items": [
                {
                    "title": "POS",
                    "route": "batch.pos",
                    "perm": "batch.pos.list",
                },
            ],
        }

    def make_admin_menu(self, request, **kwargs):
        kwargs.setdefault("include_people", False)
        menu = super().make_admin_menu(request, **kwargs)

        menu["items"] = [
            {
                "title": "Stores",
                "route": "stores",
                "perm": "stores.list",
            },
            {
                "title": "Terminals",
                "route": "terminals",
                "perm": "terminals.list",
            },
            {
                "title": "Tenders",
                "route": "tenders",
                "perm": "tenders.list",
            },
            {
                "title": "Taxes",
                "route": "taxes",
                "perm": "taxes.list",
            },
            {"type": "sep"},
        ] + menu["items"]

        return menu



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Master view for Products
"""

from wuttaweb.views import MasterView
from wuttaweb.forms.schema import WuttaQuantity, WuttaMoney

from wuttapos.db.model.products import Product
from wuttapos.server.forms.schema import DepartmentRef


class ProductView(MasterView):
    """
    Master view for Products
    """

    model_class = Product
    model_title = "Product"
    model_title_plural = "Products"

    route_prefix = "products"
    url_prefix = "/products"

    creatable = True
    editable = True
    deletable = True

    labels = {
        "product_id": "Product ID",
    }

    grid_columns = [
        "product_id",
        "brand_name",
        "description",
        "size",
        "sold_by_weight",
        "case_size",
        "department",
        "special_order",
        "unit_price_reg",
    ]

    form_fields = [
        "product_id",
        "brand_name",
        "description",
        "size",
        "department",
        "sold_by_weight",
        "case_size",
        "special_order",
        "unit_cost",
        "unit_price_reg",
        "notes",
        "on_hand",
        "on_order",
    ]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # links
        g.set_link("product_id")
        g.set_link("brand_name")
        g.set_link("description")
        g.set_link("size")

        # currency
        g.set_renderer("unit_cost", "currency", scale=4)
        g.set_renderer("unit_price_reg", "currency")

    def configure_form(self, form):
        f = form
        super().configure_form(f)
        product = f.model_instance

        # department
        f.set_node("department", DepartmentRef(self.request))

        # case_size
        f.set_node("case_size", WuttaQuantity(self.request))

        # unit_cost
        f.set_node("unit_cost", WuttaMoney(self.request, scale=4))

        # unit_price_reg
        f.set_node("unit_price_reg", WuttaMoney(self.request))

        # notes
        f.set_widget("notes", "notes")

        # on_hand
        f.set_node("on_hand", WuttaQuantity(self.request))
        if self.creating or self.editing:
            f.remove("on_hand")
        else:
            f.set_default(
                "on_hand",
                product.inventory.on_hand if product.inventory else None,
            )

        # on_order
        f.set_node("on_order", WuttaQuantity(self.request))
        if self.creating or self.editing:
            f.remove("on_order")
        else:
            f.set_default(
                "on_order",
                product.inventory.on_order if product.inventory else None,
            )


def defaults(config, **kwargs):
    base = globals()

    ProductView = kwargs.get("ProductView", base["ProductView"])
    ProductView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
"""
Master view for Stores
"""

from wuttapos.db.model.stores import Store

from wuttaweb.views import MasterView


class StoreView(MasterView):
    """
    Master view for Stores
    """

    model_class = Store
    model_title = "Store"
    model_title_plural = "Stores"

    route_prefix = "stores"
    url_prefix = "/stores"

    creatable = True
    editable = True
    deletable = True

    labels = {
        "store_id": "Store ID",
    }

    grid_columns = [
        "store_id",
        "name",
        "active",
    ]

    form_fields = [
        "store_id",
        "name",
        "active",
    ]


def defaults(config, **kwargs):
    base = globals()

    StoreView = kwargs.get("StoreView", base["StoreView"])
    StoreView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
"""
Master view for Terminals
"""

from wuttapos.db.model.terminals import Terminal

from wuttaweb.views import MasterView


class TerminalView(MasterView):
    """
    Master view for Terminals
    """

    model_class = Terminal
    model_title = "Terminal"
    model_title_plural = "Terminals"

    route_prefix = "terminals"
    url_prefix = "/terminals"

    creatable = True
    editable = True
    deletable = True

    labels = {
        "terminal_id": "Terminal ID",
    }

    grid_columns = [
        "terminal_id",
        "name",
    ]

    form_fields = [
        "terminal_id",
        "name",
    ]


def defaults(config, **kwargs):
    base = globals()

    TerminalView = kwargs.get("TerminalView", base["TerminalView"])
    TerminalView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - item lookup for department
"""

from .itemlookup import WuttaProductLookup


class WuttaProductLookupByDepartment(WuttaProductLookup):

    def __init__(self, config, department, *args, **kwargs):

        # nb. this forces first query
        kwargs.setdefault("initial_search", True)

        kwargs.setdefault("show_search", False)

        super().__init__(config, *args, **kwargs)
        model = self.app.model

        if isinstance(department, model.Department):
            self.department_key = department.uuid
        else:
            self.department_key = department

    # TODO: should somehow combine these 2 approaches, so the user can
    # still filter items within a department

    # def get_results(self, session, entry):
    #     return self.app.get_products_handler().search_products(session, entry)

    def get_results(self, session, entry):
        org = self.app.get_org_handler()
        prod = self.app.get_products_handler()
        model = self.app.model

        department = org.get_department(session, self.department_key)
        if not department:
            raise ValueError(f"department not found: {self.department_key}")

        products = (
            session.query(model.Product)
            .filter(model.Product.department == department)
            .all()
        )

        products = [prod.normalize_product(p) for p in products]

        return products



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS app
"""

from wuttjamaican import app as base


class WuttaPosAppHandler(base.AppHandler):
    """
    Custom :term:`app handler` for WuttaPOS.
    """

    default_people_handler_spec = "wuttapos.people:PeopleHandler"
    default_employment_handler_spec = "wuttapos.employment:EmploymentHandler"
    default_clientele_handler_spec = "wuttapos.clientele:ClienteleHandler"
    default_products_handler_spec = "wuttapos.products:ProductsHandler"
    default_install_handler_spec = "wuttapos.install:InstallHandler"

    def get_clientele_handler(self):
        """
        Get the configured "clientele" :term:`handler`.

        :rtype: :class:`~wuttapos.clientele.ClienteleHandler`
        """
        if "clientele" not in self.app.handlers:
            spec = self.config.get(
                f"{self.appname}.clientele.handler",
                default=self.default_clientele_handler_spec,
            )
            factory = self.app.load_object(spec)
            self.app.handlers["clientele"] = factory(self.config)
        return self.app.handlers["clientele"]

    def get_employment_handler(self):
        """
        Get the configured "employment" :term:`handler`.

        :rtype: :class:`~wuttapos.employment.EmploymentHandler`
        """
        if "employment" not in self.app.handlers:
            spec = self.config.get(
                f"{self.appname}.employment.handler",
                default=self.default_employment_handler_spec,
            )
            factory = self.app.load_object(spec)
            self.app.handlers["employment"] = factory(self.config)
        return self.app.handlers["employment"]

    def get_products_handler(self):
        """
        Get the configured "products" :term:`handler`.

        :rtype: :class:`~wuttapos.products.ProductsHandler`
        """
        if "products" not in self.app.handlers:
            spec = self.config.get(
                f"{self.appname}.products.handler",
                default=self.default_products_handler_spec,
            )
            factory = self.app.load_object(spec)
            self.app.handlers["products"] = factory(self.config)
        return self.app.handlers["products"]

    def get_employee(self, obj):
        """
        Convenience method to locate a
        :class:`~wuttjamaican.db.model.base.Person` for the given
        object.

        This delegates to the "people" handler method,
        :meth:`~wuttjamaican.people.PeopleHandler.get_person()`.
        """
        return self.get_employment_handler().get_employee(obj)


class WuttaPosAppProvider(base.AppProvider):
    """
    Custom :term:`app provider` for WuttaPOS.
    """

    email_templates = ["wuttapos:email-templates"]



"""add Tenders

Revision ID: b026e4f5c5bc
Revises: 8ce8b14af66d
Create Date: 2026-01-02 11:34:27.523125

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "b026e4f5c5bc"
down_revision: Union[str, None] = "8ce8b14af66d"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # tender
    op.create_table(
        "tender",
        sa.Column("uuid", wuttjamaican.db.util.UUID(), nullable=False),
        sa.Column("tender_id", sa.String(length=20), nullable=False),
        sa.Column("name", sa.String(length=50), nullable=False),
        sa.Column("notes", sa.Text(), nullable=True),
        sa.Column("is_cash", sa.Boolean(), nullable=False),
        sa.Column("is_foodstamp", sa.Boolean(), nullable=False),
        sa.Column("allow_cashback", sa.Boolean(), nullable=False),
        sa.Column("kick_drawer", sa.Boolean(), nullable=False),
        sa.Column("active", sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint("uuid", name=op.f("pk_tender")),
    )
    op.create_table(
        "tender_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "tender_id", sa.String(length=20), autoincrement=False, nullable=True
        ),
        sa.Column("name", sa.String(length=50), autoincrement=False, nullable=True),
        sa.Column("notes", sa.Text(), autoincrement=False, nullable=True),
        sa.Column("is_cash", sa.Boolean(), autoincrement=False, nullable=True),
        sa.Column("is_foodstamp", sa.Boolean(), autoincrement=False, nullable=True),
        sa.Column("allow_cashback", sa.Boolean(), autoincrement=False, nullable=True),
        sa.Column("kick_drawer", sa.Boolean(), autoincrement=False, nullable=True),
        sa.Column("active", sa.Boolean(), autoincrement=False, nullable=True),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_tender_version")
        ),
    )
    op.create_index(
        op.f("ix_tender_version_end_transaction_id"),
        "tender_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_tender_version_operation_type"),
        "tender_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        op.f("ix_tender_version_transaction_id"),
        "tender_version",
        ["transaction_id"],
        unique=False,
    )


def downgrade() -> None:

    # tender
    op.drop_index(op.f("ix_tender_version_transaction_id"), table_name="tender_version")
    op.drop_index(op.f("ix_tender_version_operation_type"), table_name="tender_version")
    op.drop_index(
        op.f("ix_tender_version_end_transaction_id"), table_name="tender_version"
    )
    op.drop_table("tender_version")
    op.drop_table("tender")



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Point of Sale system based on Wutta Framework
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - department lookup control
"""

from .lookup import WuttaLookup


class WuttaDepartmentLookup(WuttaLookup):

    def __init__(self, *args, **kwargs):

        # nb. this forces first query
        kwargs.setdefault("initial_search", "")
        kwargs.setdefault("allow_empty_query", True)

        super().__init__(*args, **kwargs)

    def get_results_columns(self):
        return [
            "Department ID",
            "Name",
        ]

    def get_results(self, session, entry):
        model = self.app.model
        query = session.query(model.Department).order_by(model.Department.name)

        if entry:
            query = query.filter(model.Department.name.ilike(f"%{entry}%"))

        departments = []
        for dept in query:
            departments.append(
                {
                    "uuid": dept.uuid,
                    "department_id": dept.department_id,
                    "name": dept.name,
                }
            )
        return departments

    def make_result_row(self, dept):
        return [
            dept["department_id"],
            dept["name"],
        ]



# -*- coding: utf-8; -*-
"""
Master view for Customers
"""

from wuttaweb.views import MasterView
from wuttaweb.forms.schema import PersonRef

from wuttapos.db.model.customers import Customer


class CustomerView(MasterView):
    """
    Master view for Customers
    """

    model_class = Customer
    model_title = "Customer"
    model_title_plural = "Customers"

    route_prefix = "customers"
    url_prefix = "/customers"

    creatable = True
    editable = True
    deletable = True

    labels = {
        "customer_id": "Customer ID",
    }

    grid_columns = [
        "customer_id",
        "name",
        "phone_number",
        "email_address",
    ]

    form_fields = [
        "customer_id",
        "name",
        "account_holder",
        "phone_number",
        "email_address",
    ]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # links
        g.set_link("customer_id")
        g.set_link("name")

    def configure_form(self, form):
        f = form
        super().configure_form(f)

        # account_holder
        f.set_node("account_holder", PersonRef(self.request))
        f.set_required("account_holder", False)


def defaults(config, **kwargs):
    base = globals()

    CustomerView = kwargs.get("CustomerView", base["CustomerView"])
    CustomerView.defaults(config)


def includeme(config):
    defaults(config)
